"""
This QmeQ package contains modules, where different approximate master equations
are implemented. Modules starting with `c_` are implemented using Cython.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function



"""Module for indexing many-body states using Lin tables."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import itertools

import numpy as np
from scipy.misc import factorial

from .mytypes import boolnp
from .mytypes import longnp

def binarylist_to_integer(lst):
    """
    Convert a binary number in a form of a list to an integer.

    Parameters
    ----------
    lst : list
        List containing digits of a binary number.

    Returns
    -------
    int
        A decimal integer.
    """
    return int(''.join(map(str, lst)), 2)

def integer_to_binarylist(num, binlen=0, strq=False):
    """
    Convert an integer number to a binary number in a form of a list.

    Parameters
    ----------
    num : int
        Integer number.
    binlen : int
        Length of a list containing digits of a binary number.
    strq : bool
        If true then returns a string instead of a list.

    Returns
    -------
    list
        A list containing digits of a binary number.
    """
    value = format(num, '0'+str(binlen)+'b')
    rez = value if strq else list(map(int, str(value)))
    return rez

def construct_chargelst(nsingle):
    """
    Makes list of lists containing Lin indices of the states for given charge.

    Parameters
    ----------
    nsingle : int
        Number of single particle states.

    Returns
    -------
    chargelst : list of lists
        chargelst[charge] gives a list of state indices for given charge,
        chargelst[charge][ind] gives state index.
    """
    nmany = np.power(2, nsingle)
    chargelst = [[] for i in range(nsingle+1)]
    # Iterate over many-body states
    for j1 in range(nmany):
        state = integer_to_binarylist(j1, nsingle)
        chargelst[sum(state)].append(j1)
    return chargelst

def sz_to_ind(sz, charge, nsingle):
    """
    Converts :math:`S_{z}` to a list index.

    Parameters
    ----------
    sz : int
        Value :math:`S_{z}` of a spin projection in the z direction.
    charge : int
        Value of the charge.
    nsingle : int
        Number of single particle states.

    Returns
    -------
    int
        Value of the list index corresponding to sz.
    """
    szmax = min(charge, nsingle-charge)
    return int((szmax+sz)/2)

def szrange(charge, nsingle):
    """
    Make a list giving :math:`S_{z}` values for given charge.

    Parameters
    ----------
    charge : int
        Value of the charge.
    nsingle : int
        Number of single particle states.

    Returns
    -------
    list
        List containing :math:`S_{z}` values for given charge.
    """
    szmax = min(charge, nsingle-charge)
    return list(range(-szmax, szmax+1, +2))
    #return range(szmax, -szmax-1, -2)

def empty_szlst(nsingle, noneq=False):
    """
    Make an empty list of lists corresponding to different charges and :math:`S_{z}` values.

    Parameters
    ----------
    nsingle : int
        Number of single particle states.
    noneq : bool
        If True the list contains None objects.
        If False the list contains empty lists [].

    Returns
    -------
    list of lists
        Contains None objects or [].
    """
    ncharge = nsingle+1
    if noneq:
        return [[None for j in szrange(i, nsingle)] for i in range(ncharge)]
    else:
        return [[[] for j in szrange(i, nsingle)] for i in range(ncharge)]

def construct_szlst(nsingle):
    """
    Makes list of lists of lists containing Lin indices of the states for given charge
    and spin :math:`S_{z}`.

    Parameters
    ----------
    nsingle : int
        Number of single particle states.

    Returns
    -------
    szlst : list of lists of lists
        szlst[charge] gives a list of lists of state indices for given charge,
        szlst[charge][sz] gives a list corresponding to charge and sz,
        szlst[charge][sz][ind] gives state index.
    """
    nmany = np.power(2, nsingle)
    ncharge = nsingle+1
    szlst = empty_szlst(nsingle)
    # Iterate over many-body states
    for j1 in range(nmany):
        state = integer_to_binarylist(j1, nsingle)
        charge = sum(state)
        sz = sum(state[0:int(nsingle/2)])-sum(state[int(nsingle/2):nsingle])
        szind = sz_to_ind(sz, charge, nsingle)
        szlst[charge][szind].append(j1)
    return szlst

def ssq_to_ind(ssq, sz):
    """
    Convert the value of :math:`S^{2}` for a given :math:`S_{z}` to an index.

    Parameters
    ----------
    ssq : int
        Value of :math:`S^{2}`.
    sz : int
        Value of :math:`S_{z}`.

    Returns
    -------
    int
        An index corresponding to given :math:`S^{2}` and :math:`S_{z}`.
    """
    return int((ssq-abs(sz))/2)

def ssqrange(charge, sz, nsingle):
    """
    Make a list giving all possible :math:`S^{2}` values for given charge and :math:`S_{z}`.

    Parameters
    ----------
    charge : int
        Value of the charge.
    sz : int
        Value of sz.
    nsingle : int
        Number of single particle states.

    Returns
    -------
    list
        List of all possible :math:`S^{2}` values for given charge and :math:`S_{z}`.
    """
    szmax = min(charge, nsingle-charge)
    return list(range(abs(sz), szmax+1, +2))

def empty_ssqlst(nsingle, noneq=False):
    """
    Make an empty list of lists of lists corresponding to
    different charge, :math:`S_{z}`, :math:`S^{2}` values.

    Parameters
    ----------
    nsingle : int
        Number of single particle states.
    noneq : bool
        If True the list contains None objects.
        If False the list contains empty lists [].

    Returns
    -------
    list of lists of lists
        Contains None objects or [].
    """
    ncharge = nsingle+1
    if noneq:
        return [[[None for k in ssqrange(i, j, nsingle)] for j in szrange(i, nsingle)] for i in range(ncharge)]
    else:
        return [[[[] for k in ssqrange(i, j, nsingle)] for j in szrange(i, nsingle)] for i in range(ncharge)]

def construct_ssqlst(szlst, nsingle):
    """
    Makes list of lists of lists of lists containing Lin indices of the states
    for given charge, spin projection :math:`S_{z}`, and :math:`S^{2}` value.

    Parameters
    ----------
    szlst : list of lists
        List cotaining the possbile values of :math:`S_{z}` for given charge.
        szlst[charge][szind] is an integer corresponding to :math:`S_{z}`.
    nsingle : int
        Number of single particle states.

    Returns
    -------
    sqqlst : list of lists of lists of lists
        ssqlst[charge] gives a list of lists of lists of state indices for given charge,
        ssqlst[charge][sz] gives a list of lists corresponding to charge and :math:`S_{z}`,
        ssqlst[charge][sz][ssq] gives a list corresponding to charge, :math:`S_{z}`, and :math:`S^{2}`.
        ssqlst[charge][sz][ssq][ind] gives a state index.
    """
    nmany = np.power(2, nsingle)
    ncharge = nsingle+1
    ssqlst = empty_ssqlst(nsingle)
    ssqcount = [[0 for j in ssqrange(i, i%2, nsingle)] for i in range(ncharge)]
    # Find the number of multiplets for given charge
    for j1 in range(ncharge):
        ssqr = ssqrange(j1, j1%2, nsingle)
        lssqr = len(ssqr)
        for j2 in range(lssqr):
            ssqind = ssq_to_ind(ssqr[lssqr-j2-1], 0)
            ssqcount[j1][ssqind] = len(szlst[j1][j2]) - (len(szlst[j1][j2-1]) if j2>0 else 0)
    # For Lin index charge, sz, and ssq are assigned
    ind = 0
    for charge in range(ncharge):
        for sz in szrange(charge, nsingle):
            szind = sz_to_ind(sz, charge, nsingle)
            for ssq in ssqrange(charge, sz, nsingle):
                ssqind = ssq_to_ind(ssq, sz)
                ssqind2 = ssq_to_ind(ssq, 0)
                for j1 in range(ssqcount[charge][ssqind2]):
                    ssqlst[charge][szind][ssqind].append(ind)
                    ind += 1
    return ssqlst

def flatten(lst):
    """
    Flattens one level of a list.

    Parameters
    ----------
    lst : list

    Returns
    -------
    list
        List flattened by one level.
    """
    return list(itertools.chain.from_iterable(lst))

def enum_chargelst(chargelst_lin):
    """
    Make a list of integers from 0 to 2^nsingle having nesting of chargelst_lin.

    Parameters
    ----------
    chargelst_lin : list of lists
        Constructed by construct_chargelst(nsingle).

    Returns
    -------
    chargelst : list of lists
    """
    ncharge = len(chargelst_lin)
    chargelst = [[] for i in range(ncharge)]
    counter1 = 0
    for j1 in range(ncharge):
        counter2 = counter1 + len(chargelst_lin[j1])
        chargelst[j1] = list(range(counter1, counter2))
        counter1 = counter2
    return chargelst

def enum_szlst(szlst_lin):
    """
    Make a list of integers from 0 to 2^nsingle having nesting of szlst_lin.

    Parameters
    ----------
    szlst_lin : list of lists of lists
        Constructed by construct_szlst(nsingle).

    Returns
    -------
    szlst : list of lists of lists
    """
    ncharge = len(szlst_lin)
    nsingle = ncharge-1
    szlst = empty_szlst(nsingle)
    #[[[] for j in szrange(i, nsingle)] for i in range(ncharge)]
    counter1 = 0
    for j1 in range(ncharge):
        for j2 in range(len(szlst_lin[j1])):
            counter2 = counter1 + len(szlst_lin[j1][j2])
            szlst[j1][j2] = list(range(counter1, counter2))
            counter1 = counter2
    return szlst

def make_inverse_map(lst):
    """
    Make a map from Lin index to a state index.

    Parameters
    ----------
    lst : list
        List containing Lin indices.

    Returns
    -------
    rez : list
        List containing a map from Lin index to a state index.
    """
    rez = [0]*len(lst)
    for j1 in range(len(lst)):
        rez[lst[j1]] = j1
    return rez

def make_quantum_numbers(si):
    """
    Make dictionaries between the state indices and
    quantum numbers corresponding to the states.

    Parameters
    ----------
    si : StateIndexing
        StateIndexing or StateIndexingDM object.

    Returns
    -------
    qn_ind : dict
        Dictionary from quantum number to a state.
    ind_qn : dict
        Dictionary from state index to quantum numbers of the state.
    """
    ncharge = si.ncharge
    nsingle = si.nsingle
    ind = 0
    qn_ind = {}
    ind_qn = {}
    if si.indexing == 'ssq':
        for charge in range(ncharge):
            for sz in szrange(charge, nsingle):
                szind = sz_to_ind(sz, charge, nsingle)
                for ssq in ssqrange(charge, sz, nsingle):
                    ssqind = ssq_to_ind(ssq, sz)
                    for alpha in range(len(si.ssqlst[charge][szind][ssqind])):
                        ind = si.ssqlst[charge][szind][ssqind][alpha]
                        qn_ind.update({(charge, sz, ssq, alpha):ind})
                        ind_qn.update({ind:(charge, sz, ssq, alpha)})
                        ind += 1
    elif si.indexing == 'sz':
        for charge in range(ncharge):
            for sz in szrange(charge, nsingle):
                szind = sz_to_ind(sz, charge, nsingle)
                for alpha in range(len(si.szlst[charge][szind])):
                    ind = si.szlst[charge][szind][alpha]
                    qn_ind.update({(charge, sz, alpha):ind})
                    ind_qn.update({ind:(charge, sz, alpha)})
    else:
        for charge in range(ncharge):
            for alpha in range(len(si.chargelst[charge])):
                ind = si.chargelst[charge][alpha]
                qn_ind.update({(charge, alpha):ind})
                ind_qn.update({ind:(charge, alpha)})
    return qn_ind, ind_qn

class StateIndexing(object):
    """
    Class for indexing many-body states.

    Attributes
    ----------
    i, j : list
        List of integers containing indexing.
    indexing : str
        String determining type of the indexing. Possible values are 'Lin', 'charge', 'sz', 'ssq'.
        Note that 'sz' indexing for Fock states is used for 'ssq' indexing, with
        additional specification of eigensates in Fock basis.
    symmetry : str
            String determining if the states will be augmented by a symmetry.
            Possible value is 'spin'.
    ncharge : int
        Number of charge states.
    nmany : int
        Number of many-body states.
    nsingle : int
        Number of single-particle states.
    nleads : int
        Number of leads. Set to zero by default.
    chargelst_lin : list of lists
        chargelst_lin[charge] gives a list of state Lin indices for given charge,
        chargelst_lin[charge][ind] gives state Lin index
    szlst_lin : list of lists
        szlst_lin[charge] gives a list of lists of state indices for given charge,
        szlst_lin[charge][sz] gives a list corresponding to charge and sz,
        szlst_lin[charge][sz][ind] gives state Lin index
    chargelst : list of lists
        Indices of states for different charges for chosen indexing.
    szlst : list of lists
        Indices of states for different :math:`S_{z}` values for chosen indexing.
    ssqlst : list of lists of lists
        Indices of states for different :math:`S_{z}` and :math:`S^{2}` values.
    qn_ind : dictionary
        Maps quantum number to state index
    ind_qn : dictionary
        Maps state index to quantum numbers
    states_order : list
        Order of states used by many-body state printing functions in qmeq.various
        print_state() and print_all_states()
    """

    def __init__(self, nsingle, indexing='Lin', symmetry='n', nleads=0):
        """
        Initialization of the StateIndexing class

        Parameters
        ----------
        nsingle : int
            Number of single-particle states.
        indexing : str
            String determining type of the indexing. Possible values are 'Lin', 'charge', 'sz', 'ssq'.
        symmetry : str
            String determining that the states will be augmented by the symmetry.
            Possible value is 'spin'.
        """
        #
        self.nsingle_sym = nsingle//2 if symmetry is 'spin' else nsingle
        self.nsingle = nsingle
        self.indexing = indexing
        self.symmetry = symmetry
        self.ncharge = nsingle+1
        self.nmany = 2**nsingle
        self.nleads = nleads
        self.nleads_sym = nleads//2 if symmetry is 'spin' else nleads
        #
        self.szlst_lin = None
        self.szlst = None
        self.ssqlst = None
        if indexing == 'charge':
            self.chargelst_lin = construct_chargelst(nsingle)
            self.chargelst = enum_chargelst(self.chargelst_lin)
            self.i = flatten(self.chargelst_lin)
        elif (indexing == 'sz' or indexing == 'ssq') and self.nsingle%2 == 0:
            self.chargelst_lin = construct_chargelst(nsingle)
            self.chargelst = enum_chargelst(self.chargelst_lin)
            self.szlst_lin = construct_szlst(self.nsingle)
            self.szlst = enum_szlst(self.szlst_lin)
            self.i = flatten(flatten(self.szlst_lin))
            if indexing == 'ssq':
                self.ssqlst = construct_ssqlst(self.szlst, nsingle)
        elif (indexing == 'sz' or indexing == 'ssq') and self.nsingle%2 != 0:
            print("WARNING: For 'sz' or 'ssq' indexing, nsingle has to be even. \
                   Using 'Lin' indexing.")
            self.indexing = 'Lin'
            self.chargelst = construct_chargelst(nsingle)
            self.i = list(range(self.nmany))
        elif indexing == 'Lin':
            self.chargelst = construct_chargelst(nsingle)
            self.i = list(range(self.nmany))
        else:
            print("WARNING: The indexing has to be 'Lin', 'charge', or 'sz'. \
                   Using 'Lin' indexing.")
            self.indexing = 'Lin'
            self.chargelst = construct_chargelst(nsingle)
            self.i = list(range(self.nmany))
        self.j = make_inverse_map(self.i)
        # Note that these quantum numbers to state and state to quantum numbers dictionaries
        # are necessary only for ssq indexing
        self.qn_ind, self.ind_qn = make_quantum_numbers(self)
        self.states_order = list(range(self.nmany))

    def get_state(self, ind, linq=False, strq=False):
        """
        Returns a list containing digits of a binary number corresponding to a state index.

        Parameters
        ----------
        ind : int
            Index of the state.
        linq : bool
            For linq=True uses Lin indexing, for linq=False uses specified indexing.
        strq : bool
            If true then returns a string instead of a list.

        Returns
        -------
        list
            List containing digits of a binary number corresponding to a state index.
        """
        if linq:
            return integer_to_binarylist(ind, self.nsingle, strq)
        else:
            return integer_to_binarylist(self.i[ind], self.nsingle, strq)

    def get_ind(self, state, linq=False):
        """
        Returns an index for given state

        Parameters
        ----------
        state : list
           List containing digits of a binary number.
        linq : bool
            For linq=True uses Lin indexing, for linq=False uses specified indexing

        Returns
        -------
        int
            Index of the state.
        """
        if linq:
            return binarylist_to_integer(state)
        else:
            return self.j[binarylist_to_integer(state)]

    def get_lst(self, charge='n', sz='n', ssq='n'):
        """
        Gives a list of state Lin indices corresponding to charge or charge and sz.

        Parameters
        ----------
        charge : int
            Value of charge
        sz : int
            Value of sz
        ssq : int
            Value of total angular momentum ssq

        Returns
        -------
        list
            List of indices corresponding to given charge, chargelst[charge],
            or given charge and sz, szlst[charge][szind],
            or given charge, sz, and, ssq, szlst[charge][szind][ssqind].
        """
        if charge == 'n' and sz == 'n':
            return None
        elif sz == 'n':
            return self.chargelst[charge]
        elif self.szlst != None and ssq == 'n':
            szind = sz_to_ind(sz, charge, self.nsingle)
            return self.szlst[charge][szind]
        elif self.szlst != None and self.ssqlst != None:
            szind = sz_to_ind(sz, charge, self.nsingle)
            ssqind = ssq_to_ind(ssq, sz)
            return self.ssqlst[charge][szind][ssqind]
        else:
            print("WARNING: No indexing by 'sz' or 'ssq'. Returning charge list.")
            return self.chargelst[charge]

class StateIndexingPauli(StateIndexing):
    """
    Class for indexing diagonal density matrix elements (Pauli master equation).
    Derived from StateIndexing.

    Attributes
    ----------
    statesdm : list
        List of states considered in calculations involving density matrix.
    shiftlst0 : numpy array
        Array necessary for labeling states for Pauli master equation. See method get_ind_dm0().
    dictdm : numpy array
        Array giving enumeration of considered many-body states in statesdm.
    mapdm0 : list
        List showing which states are mapped to each other due to symmetries.
        For example mapdm0[1]=1, mapdm0[2]=1, shows that 1 and 2 are equivalent.
        Also mapdm0[ind]=-1 shows that the state is excluded.
    booldm0 : list
        List giving states which will be used. Other states will be mapped to states,
        which have booldm0[ind]=True.
        From example for mapdm0 we have booldm0[1]=True, booldm0[2]=False.
    """

    def __init__(self, nsingle, indexing='Lin', symmetry='n', nleads=0):
        """
        Initialization of the StateIndexingDM class

        Parameters
        ----------
        nsingle : int
            Number of single-particle states.
        indexing : str
            String determining type of the indexing. Possible values are 'Lin', 'charge', 'sz', 'ssq'.
        symmetry : str
            String determining that the states will be augmented by the symmetry.
        """
        StateIndexing.__init__(self, nsingle, indexing, symmetry, nleads)
        self.shiftlst0 = np.zeros(self.ncharge+1, dtype=longnp)
        self.dictdm = np.zeros(self.nmany, dtype=longnp)
        self.statesdm = None
        self.mapdm0, self.booldm0 = None, None
        self.set_statesdm(self.chargelst)

    def set_statesdm(self, statesdm):
        """
        Sets the considered many-body states for Pauli master equation calculations.

        Parameters
        ----------
        statesdm: list
            List containing indices of many-body state under consideration.
        """
        self.statesdm = statesdm
        self.statesdm.append([])
        self.npauli_ = 0
        for j1 in range(self.ncharge):
            self.npauli_ += len(statesdm[j1])
        self.set_dictdm()

    def set_dictdm(self):
        """
        Makes dictdm, shiftlst0 for Pauli master equation indexing.
        """
        for j1 in range(self.ncharge):
            self.shiftlst0[j1+1] = self.shiftlst0[j1] + len(self.statesdm[j1])
            counter = 0
            for j2 in self.statesdm[j1]:
                self.dictdm[j2] = counter
                counter += 1
        self.set_mapdm()

    def set_mapdm(self):
        """
        Reduce the number of diagonal matrix elements by using symmetries.
        """
        #
        def add_elem(counter, b, bp, charge, dictq=True):
            bbp = self.get_ind_dm0(b, bp, charge, maptype=0)
            self.mapdm0[bbp] = counter
            self.booldm0[bbp] = dictq
            #if dictq:
            #    self.inddm0.update({counter:(b, bp)})
            return counter+1
        #
        self.mapdm0 = np.ones(self.npauli_, dtype=longnp)*(-1)
        self.booldm0 = np.zeros(self.npauli_, dtype=boolnp)
        #self.inddm0 = {}
        counter = 0
        # Diagonal density matrix elements
        for charge in range(self.ncharge):
            for b in self.statesdm[charge]:
                if self.indexing == 'ssq':
                    (b_ch,  b_sz,  b_ssq,  b_alpha)  = self.ind_qn[b]
                    if b_ssq == -b_sz:
                        add_elem(counter, b, b, charge)
                        for sz in range(b_sz+2, b_ssq+1, 2):
                            b1  = self.qn_ind[(b_ch,  sz, b_ssq,  b_alpha)]
                            add_elem(counter, b1, b1, charge, dictq=False)
                        counter = counter+1
                else: counter = add_elem(counter, b, b, charge)
        self.npauli = counter

    def get_ind_dm0(self, b, bp, charge, maptype=1):
        """
        Get the state index for Pauli master equation.

        Parameters
        ----------
        b,bp : int
            Indices of many-body states within the same charge state.
            Note that bp is just a dummy variable for Pauli master equation.
        charge : int
            Charge of b and bp.

        Returns
        -------
        int
            Index of the zeroth order density matrix element.
        """
        if maptype == 0:
            return self.dictdm[b] + self.shiftlst0[charge]
        elif maptype == 1:
            return self.mapdm0[self.dictdm[b] + self.shiftlst0[charge]]
        elif maptype == 2:
            return self.booldm0[self.dictdm[b] + self.shiftlst0[charge]]

class StateIndexingDM(StateIndexing):
    """
    Class for indexing density matrix elements.
    Separates the off-diagonal density matrix elements into real and imaginary parts.
    Derived from StateIndexing.

    Attributes
    ----------
    ndm0 : int
        Number of density matrix elements to zeroth order corresponding to statesdm,
        when symmetries are used.
    ndm0r : int
        Number of unique density matrix elements to zeroth order corresponding to statesdm,
        when symmetries are used.
        The real and imaginary parts of the off-diagonal matrix element are considered
        as separate entities.
    ndm0\_ : int
        Number of density matrix elements to zeroth order corresponding to statesdm,
        when no symmetries are used.
    ndm0_tot : int
        Total number of density matrix elements to zeroth order for given nsingle,
        when no symmetries are used.
    ndm1, ndm1\_ : int, int
        Number of density matrix elements to first order corresponding to statesdm,
        when no symmetries are used.
    ndm1_tot : int
        Total number of density matrix elements to first order for given nsingle,
        when no symmetries are used.
    npauli\_ : int
        Number of diagonal density matrix elements to zeroth order corresponding to statesdm,
        when no symmetries are used.
    npauli : int
        Number of diagonal density matrix elements to zeroth order corresponding to statesdm,
        when symmetries are used.
    statesdm : list
        List of states considered in calculations involving density matrix.
    shiftlst0 : numpy array
        Array necessary for labeling zeroth order density matrix elements. See method get_ind_dm0().
    shiftlst1 : numpy array
        Array necessary for labeling first order density matrix elements. See method get_ind_dm1().
    lenlst : numpy array
        Array containing number of many-body states for each charge. lenlst[charge]=number of states.
    dictdm : numpy array
        Array giving enumeration of considered many-body states in statesdm.
    inddm0 : dict
        Dictionary between index of density matrix element and pair of many-body states (a,b).
    mapdm0 : list
        List showing which density matrix elements are mapped to each other due to symmetries.
        For example mapdm0[1]=1, mapdm0[2]=1, shows that 1 and 2 are equivalent.
        Also mapdm0[ind]=-1 shows that the state is excluded.
    booldm0 : list
        List giving density matrix elements which will be used.
        Other elements will be mapped to elements, which have booldm0[ind]=True.
        From example for mapdm0 we have booldm0[1]=True, booldm0[2]=False.
    conjdm0 : list
        List showing, which density matrix elements are complex conjugate and are not unique.
    """

    def __init__(self, nsingle, indexing='Lin', symmetry='n', nleads=0):
        """
        Initialization of the StateIndexingDM class

        Parameters
        ----------
        nsingle : int
            Number of single-particle states.
        indexing : str
            String determining type of the indexing. Possible values are 'Lin', 'charge', 'sz', 'ssq'.
        symmetry : str
            String determining that the states will be augmented by the symmetry.
        """
        StateIndexing.__init__(self, nsingle, indexing, symmetry, nleads)
        self.ndm0_tot = int(factorial(2*self.nsingle)/factorial(self.nsingle)**2)
        self.ndm1_tot = int(self.nsingle/(self.nsingle+1)*self.ndm0_tot)
        #
        self.shiftlst0 = np.zeros(self.ncharge+1, dtype=longnp)
        self.shiftlst1 = np.zeros(self.ncharge, dtype=longnp)
        self.lenlst = np.zeros(self.ncharge, dtype=longnp)
        self.dictdm = np.zeros(self.nmany, dtype=longnp)
        #
        self.statesdm, self.mapdm0, self.mapdm1 = None, None, None
        self.set_statesdm(self.chargelst)

    def set_statesdm(self, statesdm):
        """
        Sets the considered many-body states for density matrix calculations.

        Parameters
        ----------
        statesdm: list
            List containing indices of many-body state under consideration.
        """
        self.statesdm = list(statesdm)
        self.statesdm.append([])
        self.ndm0_, self.ndm1_, self.npauli_ = 0, 0, 0
        for j1 in range(self.ncharge):
            self.npauli_ += len(statesdm[j1])
            self.ndm0_ += len(statesdm[j1])**2
            if j1 < self.ncharge-1:
                self.ndm1_ += len(statesdm[j1])*len(statesdm[j1+1])
        self.set_dictdm()

    def set_dictdm(self):
        """
        Makes dictdm, shiftlst0, and shiftlst1 necessary for density-matrix element indexing.
        """
        for j1 in range(self.ncharge):
            self.lenlst[j1] = len(self.statesdm[j1])
            self.shiftlst0[j1+1] = self.shiftlst0[j1] + len(self.statesdm[j1])**2
            if j1 < self.ncharge-1:
                self.shiftlst1[j1+1] = (self.shiftlst1[j1]
                                       +len(self.statesdm[j1])*len(self.statesdm[j1+1]))
            counter = 0
            for j2 in self.statesdm[j1]:
                self.dictdm[j2] = counter
                counter += 1
        self.set_mapdm()

    def set_mapdm(self):
        """
        Reduce the number of diagonal matrix elements by using symmetries.
        """
        #
        def add_elem(counter, b, bp, charge, dictq=True, conjq=True):
            bbp = self.get_ind_dm0(b, bp, charge, maptype=0)
            self.mapdm0[bbp] = counter
            self.booldm0[bbp] = dictq
            self.conjdm0[bbp] = conjq
            if dictq:
                self.inddm0.update({counter:(b, bp)})
            return counter+1
        #
        self.mapdm0 = np.ones(self.ndm0_, dtype=longnp)*(-1)
        self.booldm0 = np.zeros(self.ndm0_, dtype=boolnp)
        self.conjdm0 = np.zeros(self.ndm0_, dtype=boolnp)
        self.inddm0 = {}
        counter = 0
        # Diagonal density matrix elements
        for charge in range(self.ncharge):
            for b in self.statesdm[charge]:
                if self.indexing == 'ssq':
                    (b_ch,  b_sz,  b_ssq,  b_alpha)  = self.ind_qn[b]
                    if b_ssq == -b_sz:
                        add_elem(counter, b, b, charge)
                        for sz in range(b_sz+2, b_ssq+1, 2):
                            b1  = self.qn_ind[(b_ch,  sz, b_ssq,  b_alpha)]
                            add_elem(counter, b1, b1, charge, dictq=False)
                        counter = counter+1
                else: counter = add_elem(counter, b, b, charge)
        self.npauli = counter
        # Off-diagoanl density matrix elements
        for charge in range(self.ncharge):
            for b, bp in itertools.combinations(self.statesdm[charge], 2):
                if self.indexing == 'sz':
                    if self.ind_qn[b][1] == self.ind_qn[bp][1]:
                        add_elem(counter, bp, b, charge, dictq=False, conjq=False)
                        counter = add_elem(counter, b, bp, charge)
                elif self.indexing == 'ssq':
                    (b_ch,  b_sz,  b_ssq,  b_alpha)  = self.ind_qn[b]
                    (bp_ch, bp_sz, bp_ssq, bp_alpha) = self.ind_qn[bp]
                    if (b_sz, b_ssq) == (bp_sz, bp_ssq):
                        if b_ssq == -b_sz:
                            add_elem(counter, bp, b, charge, dictq=False, conjq=False)
                            add_elem(counter, b, bp, charge)
                            for sz in range(b_sz+2, b_ssq+1, 2):
                                b1  = self.qn_ind[(b_ch,  sz, b_ssq,  b_alpha)]
                                b1p = self.qn_ind[(bp_ch, sz, bp_ssq, bp_alpha)]
                                add_elem(counter, b1p, b1, charge, dictq=False, conjq=False)
                                add_elem(counter, b1, b1p, charge, dictq=False)
                            counter = counter+1
                else:
                    add_elem(counter, bp, b, charge, dictq=False, conjq=False)
                    counter = add_elem(counter, b, bp, charge)
        self.ndm0 = counter
        self.ndm0r = self.npauli+2*(self.ndm0-self.npauli)
        self.ndm1 = self.ndm1_

    def get_ind_dm0(self, b, bp, charge, maptype=1):
        """
        Get the index of zeroth order density matrix element.

        Parameters
        ----------
        b,bp : int
            Indices of many-body states within the same charge state.
        charge : int
            Charge of b and bp.

        Returns
        -------
        int
            Index of the zeroth order density matrix element.
        """
        # l = len(self.statesdm[charge])
        # i = self.dictdm[b]
        # j = self.dictdm[bp]
        # shiftas = self.shiftlst0[charge]
        # print('index is', l*i + j + shiftas)
        if maptype == 0:
            return (self.lenlst[charge]*self.dictdm[b] + self.dictdm[bp]
                   + self.shiftlst0[charge])
        elif maptype == 1:
            return (self.mapdm0[self.lenlst[charge]*self.dictdm[b] + self.dictdm[bp]
                   + self.shiftlst0[charge]])
        elif maptype == 2:
            return (self.booldm0[self.lenlst[charge]*self.dictdm[b] + self.dictdm[bp]
                   + self.shiftlst0[charge]])
        elif maptype == 3:
            return (self.conjdm0[self.lenlst[charge]*self.dictdm[b] + self.dictdm[bp]
                   + self.shiftlst0[charge]])

    def get_ind_dm1(self, c, b, bcharge, maptype=1):
        """
        Get the index of first order density matrix element.

        Parameters
        ----------
        c,b : int
            Indices of many-body states differing by single charge, charge(c)>charge(b).
        bcharge : int
            Charge of many-body state b.

        Returns
        -------
        int
            Index of the first order density matrix element.
        """
        # l = len(self.statesdm[charge])
        # i = self.dictdm[c]
        # j = self.dictdm[b]
        # shiftas = self.shiftlst1[charge]
        # print('index is', l*i + j + shiftas)
        #       b
        #     -----
        #   c |   |
        #     -----
        return self.lenlst[bcharge]*self.dictdm[c] + self.dictdm[b] + self.shiftlst1[bcharge]

class StateIndexingDMc(StateIndexing):
    """
    Class for indexing density matrix elements.
    Does not separates the off-diagonal density matrix elements into real and imaginary parts,
    and treats Phi[0]_{b,bp} and Phi[0]_{bp,b} as separate entitites.
    Derived from StateIndexing.

    Attributes
    ----------
    Same as in StateIndexingDM.
    """

    def __init__(self, nsingle, indexing='Lin', symmetry='n', nleads=0):
        """
        Initialization of the StateIndexingDMc class

        Parameters
        ----------
        nsingle : int
            Number of single-particle states.
        indexing : str
            String determining type of the indexing. Possible values are 'Lin', 'charge', 'sz', 'ssq'.
        symmetry : str
            String determining that the states will be augmented by the symmetry.
        """
        StateIndexing.__init__(self, nsingle, indexing, symmetry, nleads)
        self.ndm0_tot = int(factorial(2*self.nsingle)/factorial(self.nsingle)**2)
        self.ndm1_tot = int(self.nsingle/(self.nsingle+1)*self.ndm0_tot)
        #
        self.shiftlst0 = np.zeros(self.ncharge+1, dtype=longnp)
        self.shiftlst1 = np.zeros(self.ncharge, dtype=longnp)
        self.lenlst = np.zeros(self.ncharge, dtype=longnp)
        self.dictdm = np.zeros(self.nmany, dtype=longnp)
        #
        self.statesdm, self.mapdm0, self.mapdm1 = None, None, None
        self.set_statesdm(self.chargelst)

    def set_statesdm(self, statesdm):
        """
        Sets the considered many-body states for density matrix calculations.

        Parameters
        ----------
        statesdm: list
            List containing indices of many-body state under consideration.
        """
        self.statesdm = statesdm
        self.statesdm.append([])
        self.ndm0_, self.ndm1_, self.npauli_ = 0, 0, 0
        for j1 in range(self.ncharge):
            self.npauli_ += len(statesdm[j1])
            self.ndm0_ += len(statesdm[j1])**2
            if j1 < self.ncharge-1:
                self.ndm1_ += len(statesdm[j1])*len(statesdm[j1+1])
        self.set_dictdm()

    def set_dictdm(self):
        """
        Makes dictdm, shiftlst0, and shiftlst1 necessary for density-matrix element indexing.
        """
        for j1 in range(self.ncharge):
            self.lenlst[j1] = len(self.statesdm[j1])
            self.shiftlst0[j1+1] = self.shiftlst0[j1] + len(self.statesdm[j1])**2
            if j1 < self.ncharge-1:
                self.shiftlst1[j1+1] = (self.shiftlst1[j1]
                                       +len(self.statesdm[j1])*len(self.statesdm[j1+1]))
            counter = 0
            for j2 in self.statesdm[j1]:
                self.dictdm[j2] = counter
                counter += 1
        self.set_mapdm()

    def set_mapdm(self):
        """
        Reduce the number of diagonal matrix elements by using symmetries.
        """
        #
        def add_elem(counter, b, bp, charge, dictq=True):
            bbp = self.get_ind_dm0(b, bp, charge, maptype=0)
            self.mapdm0[bbp] = counter
            self.booldm0[bbp] = dictq
            if dictq:
                self.inddm0.update({counter:(b, bp)})
            return counter+1
        #
        self.mapdm0 = np.ones(self.ndm0_, dtype=longnp)*(-1)
        self.booldm0 = np.zeros(self.ndm0_, dtype=boolnp)
        self.inddm0 = {}
        counter = 0
        # Diagonal density matrix elements
        for charge in range(self.ncharge):
            for b in self.statesdm[charge]:
                if self.indexing == 'ssq':
                    (b_ch,  b_sz,  b_ssq,  b_alpha)  = self.ind_qn[b]
                    if b_ssq == -b_sz:
                        add_elem(counter, b, b, charge)
                        for sz in range(b_sz+2, b_ssq+1, 2):
                            b1  = self.qn_ind[(b_ch,  sz, b_ssq,  b_alpha)]
                            add_elem(counter, b1, b1, charge, dictq=False)
                        counter = counter+1
                else: counter = add_elem(counter, b, b, charge)
        self.npauli = counter
        #self.npauli = 0
        # Off-diagoanl density matrix elements
        for charge in range(self.ncharge):
            #for b, bp in itertools.product(self.statesdm[charge], self.statesdm[charge]):
            for b, bp in itertools.permutations(self.statesdm[charge], 2):
                if self.indexing == 'sz':
                    if self.ind_qn[b][1] == self.ind_qn[bp][1]:
                        counter = add_elem(counter, b, bp, charge)
                elif self.indexing == 'ssq':
                    (b_ch,  b_sz,  b_ssq,  b_alpha)  = self.ind_qn[b]
                    (bp_ch, bp_sz, bp_ssq, bp_alpha) = self.ind_qn[bp]
                    if (b_sz, b_ssq) == (bp_sz, bp_ssq):
                        if b_ssq == -b_sz:
                            add_elem(counter, b, bp, charge)
                            for sz in range(b_sz+2, b_ssq+1, 2):
                                b1  = self.qn_ind[(b_ch,  sz, b_ssq,  b_alpha)]
                                b1p = self.qn_ind[(bp_ch, sz, bp_ssq, bp_alpha)]
                                add_elem(counter, b1, b1p, charge, dictq=False)
                            counter = counter+1
                else: counter = add_elem(counter, b, bp, charge)
        self.ndm0 = counter
        self.ndm0r = self.npauli+2*(self.ndm0-self.npauli)
        self.ndm1 = self.ndm1_

    def get_ind_dm0(self, b, bp, charge, maptype=1):
        """
        Get the index of zeroth order density matrix element.

        Parameters
        ----------
        b,bp : int
            Indices of many-body states within the same charge state.
        charge : int
            Charge of b and bp.

        Returns
        -------
        int
            Index of the zeroth order density matrix element.
        """
        # l = len(self.statesdm[charge])
        # i = self.dictdm[b]
        # j = self.dictdm[bp]
        # shiftas = self.shiftlst0[charge]
        # print('index is', l*i + j + shiftas)
        if maptype == 0:
            return self.lenlst[charge]*self.dictdm[b] + self.dictdm[bp] + self.shiftlst0[charge]
        elif maptype == 1:
            return (self.mapdm0[self.lenlst[charge]*self.dictdm[b] + self.dictdm[bp]
                   + self.shiftlst0[charge]])
        elif maptype == 2:
            return (self.booldm0[self.lenlst[charge]*self.dictdm[b] + self.dictdm[bp]
                   + self.shiftlst0[charge]])

    def get_ind_dm1(self, c, b, bcharge, maptype=1):
        """
        Get the index of first order density matrix element.

        Parameters
        ----------
        c,b : int
            Indices of many-body states differing by single charge, charge(c)>charge(b).
        bcharge : int
            Charge of many-body state b.

        Returns
        -------
        int
            Index of the first order density matrix element.
        """
        # l = len(self.statesdm[charge])
        # i = self.dictdm[c]
        # j = self.dictdm[b]
        # shiftas = self.shiftlst1[charge]
        # print('index is', l*i + j + shiftas)
        #       b
        #     -----
        #   c |   |
        #     -----
        return self.lenlst[bcharge]*self.dictdm[c] + self.dictdm[b] + self.shiftlst1[bcharge]



"""Module containing various special functions, cython implementation."""

# cython: profile=True

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np
from scipy.special import psi as digamma
from scipy.integrate import quad

from .mytypes import doublenp
from .mytypes import complexnp

cimport numpy as np
cimport cython

# These definitions are already specified in specfuncc.pxd
# as well as 'import numpy as np' and 'cimport numpy as np'
'''
ctypedef np.uint8_t bool_t
ctypedef np.int_t int_t
ctypedef np.int64_t long_t
ctypedef np.float64_t double_t
ctypedef np.complex128_t complex_t
'''

cdef double_t pi = 3.14159265358979323846

from libc.math cimport exp
#cdef extern from "math.h":
#    double_t exp(double_t)

from libc.math cimport log
#cdef extern from "math.h":
#    double_t log(double_t)

@cython.cdivision(True)
cdef double_t fermi_func(double_t x):
    return 1./(exp(x)+1.)

@cython.boundscheck(False)
@cython.cdivision(True)
cdef int_t func_pauli(double_t Ecb, double_t mu, double_t T,
                      double_t Dm, double_t Dp, int_t itype,
                      np.ndarray[double_t, ndim=1] rez):
    cdef double_t alpha, Rm, Rp, cur0, cur1
    alpha = (Ecb-mu)/T
    Rm, Rp = (Dm-mu)/T, (Dp-mu)/T
    if itype == 1 or itype == 3 or (alpha < Rp and alpha > Rm):
        cur0 = fermi_func(alpha)
        cur1 = 1.0-cur0
        rez[0] = 2*pi*cur0
        rez[1] = 2*pi*cur1
    else:
        rez[0], rez[1] = 0.0, 0.0
    return 0

@cython.boundscheck(False)
@cython.cdivision(True)
cdef int_t func_1vN(double_t Ecb, double_t mu, double_t T,
                    double_t Dm, double_t Dp,
                    int_t itype, int_t limit,
                    np.ndarray[complex_t, ndim=1] rez):
    cdef double_t alpha, Rm, Rp, err
    cdef complex_t cur0, cur1, en0, en1, const0, const1
    #-------------------------
    if itype == 0:
        alpha, Rm, Rp = (Ecb-mu)/T, (Dm-mu)/T, (Dp-mu)/T
        cur0, err = quad(fermi_func, Rm, Rp, weight='cauchy', wvar=alpha,
                                             epsabs=1.0e-6, epsrel=1.0e-6, limit=limit)
        cur0 = cur0 + (-1.0j*pi*fermi_func(alpha) if alpha < Rp and alpha > Rm else 0.0j)
        cur1 = cur0 + log(abs((Rm-alpha)/(Rp-alpha)))
        cur1 = cur1 + (1.0j*pi if alpha < Rp and alpha > Rm else 0.0j)
        #
        const0 = T*((-Rm if Rm < -40.0 else log(1+exp(-Rm)))
                   -(-Rp if Rp < -40.0 else log(1+exp(-Rp))))
        const1 = const0 + Dm-Dp
        #
        en0 = const0 + Ecb*cur0
        en1 = const1 + Ecb*cur1
    elif itype == 1:
        alpha, Rm, Rp = (Ecb-mu)/T, Dm/T, Dp/T
        cur0 = digamma(0.5+1.0j*alpha/(2*pi)).real - log(abs(Rm)/(2*pi))
        cur0 = cur0 - 1.0j*pi*fermi_func(alpha)
        cur1 = cur0 + log(abs(Rm/Rp))
        cur1 = cur1 + 1.0j*pi
        #
        en0 = -T*Rm + Ecb*cur0
        en1 = -T*Rp + Ecb*cur1
    elif itype == 2:
        alpha, Rm, Rp = (Ecb-mu)/T, (Dm-mu)/T, (Dp-mu)/T
        cur0 = -1.0j*pi*fermi_func(alpha) if alpha < Rp and alpha > Rm else 0.0j
        cur1 = cur0 + (1.0j*pi if alpha < Rp and alpha > Rm else 0.0j)
        en0 = Ecb*cur0
        en1 = Ecb*cur1
    elif itype == 3:
        alpha = (Ecb-mu)/T
        cur0 = -1.0j*pi*fermi_func(alpha)
        cur1 = cur0 + 1.0j*pi
        en0 = Ecb*cur0
        en1 = Ecb*cur1
    #-------------------------
    rez[0], rez[1], rez[2], rez[3] = cur0, cur1, en0, en1
    return 0

def c_fermi_func(double_t x):
    return fermi_func(x)

def c_func_pauli(double_t Ecb, double_t mu, double_t T,
                 double_t Dm, double_t Dp, int_t itype):
    rez = np.zeros(2, dtype=doublenp)
    func_pauli(Ecb, mu, T, Dm, Dp, itype, rez)
    return rez

def c_func_1vN(double_t Ecb, double_t mu, double_t T,
               double_t Dm, double_t Dp,
               int_t itype, int_t limit):
    rez = np.zeros(4, dtype=complexnp)
    func_1vN(Ecb, mu, T, Dm, Dp, itype, limit, rez)
    return rez



"""Module containing python functions, which generate first order Linblad kernels."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import numpy as np
import itertools

from ..mytypes import complexnp
from ..mytypes import doublenp

from ..specfunc import func_pauli
from ..aprclass import Approach

#---------------------------------------------------------------------------------------------------
# Lindblad approach
#---------------------------------------------------------------------------------------------------
def generate_tLba(sys):
    """
    Make factors used for generating Lindblad master equation kernel.

    Parameters
    ----------
    sys : Approach
        Approach object.

    Modifies:
    sys.tLba : array
        Jump operator matrix in many-body basis.
    """
    (Tba, E, si) = (sys.leads.Tba, sys.qd.Ea, sys.si)
    (mulst, tlst, dlst) = (sys.leads.mulst, sys.leads.tlst, sys.leads.dlst)
    (mtype, itype) = (sys.leads.mtype, sys.funcp.itype)
    #
    tLba = np.zeros(Tba.shape, dtype=mtype)
    for charge in range(si.ncharge-1):
        bcharge = charge+1
        acharge = charge
        for b, a in itertools.product(si.statesdm[bcharge], si.statesdm[acharge]):
            Eba = E[b]-E[a]
            for l in range(si.nleads):
                fct1, fct2 = func_pauli(Eba, mulst[l], tlst[l], dlst[l,0], dlst[l,1], itype)
                tLba[l, b, a] = np.sqrt(fct1)*Tba[l, b, a]
                tLba[l, a, b] = np.sqrt(fct2)*Tba[l, a, b]
    sys.tLba = tLba
    return 0

def generate_kern_lindblad(sys):
    """
    Generates a kernel (Liouvillian) matrix corresponding to Lindblad approach.

    Parameters
    ----------
    sys : Approach
        Approach object.

    Modifies:
    sys.kern : array
        Kernel matrix for first-order Lindblad approach.
    sys.bvec : array
        Right hand side column vector for master equation.
        The entry funcp.norm_row is 1 representing normalization condition.
    """
    (E, tLba) = (sys.qd.Ea, sys.tLba)
    (si, symq, norm_rowp) = (sys.si, sys.funcp.symq, sys.funcp.norm_row)
    norm_row = norm_rowp if symq else si.ndm0r
    last_row = si.ndm0r-1 if symq else si.ndm0r
    kern = np.zeros((last_row+1, si.ndm0r), dtype=doublenp)
    bvec = np.zeros(last_row+1, dtype=doublenp)
    bvec[norm_row] = 1
    for charge in range(si.ncharge):
        for b, bp in itertools.combinations_with_replacement(si.statesdm[charge], 2):
            bbp = si.get_ind_dm0(b, bp, charge)
            bbp_bool = si.get_ind_dm0(b, bp, charge, 2)
            if bbp != -1 and bbp_bool:
                bbpi = si.ndm0 + bbp - si.npauli
                bbpi_bool = True if bbpi >= si.ndm0 else False
                if bbpi_bool:
                    kern[bbp, bbpi] += E[b]-E[bp]
                    kern[bbpi, bbp] += E[bp]-E[b]
                #--------------------------------------------------
                for a, ap in itertools.product(si.statesdm[charge-1], si.statesdm[charge-1]):
                    aap = si.get_ind_dm0(a, ap, charge-1)
                    if aap != -1:
                        fct_aap = 0
                        for l in range(si.nleads):
                            fct_aap += tLba[l, b, a]*tLba[l, bp, ap].conjugate()
                        aapi = si.ndm0 + aap - si.npauli
                        aap_sgn = +1 if si.get_ind_dm0(a, ap, charge-1, maptype=3) else -1
                        kern[bbp, aap] += fct_aap.real                          # kern[bbp, aap]   += fct_aap.real
                        if aapi >= si.ndm0:
                            kern[bbp, aapi] -= fct_aap.imag*aap_sgn             # kern[bbp, aapi]  -= fct_aap.imag*aap_sgn
                            if bbpi_bool:
                                kern[bbpi, aapi] += fct_aap.real*aap_sgn        # kern[bbpi, aapi] += fct_aap.real*aap_sgn
                        if bbpi_bool:
                            kern[bbpi, aap] += fct_aap.imag                     # kern[bbpi, aap]  += fct_aap.imag
                #--------------------------------------------------
                for bpp in si.statesdm[charge]:
                    bppbp = si.get_ind_dm0(bpp, bp, charge)
                    if bppbp != -1:
                        fct_bppbp = 0
                        for a in si.statesdm[charge-1]:
                            for l in range(si.nleads):
                                fct_bppbp += -0.5*tLba[l, a, b].conjugate()*tLba[l, a, bpp]
                        for c in si.statesdm[charge+1]:
                            for l in range(si.nleads):
                                fct_bppbp += -0.5*tLba[l, c, b].conjugate()*tLba[l, c, bpp]
                        bppbpi = si.ndm0 + bppbp - si.npauli
                        bppbp_sgn = +1 if si.get_ind_dm0(bpp, bp, charge, maptype=3) else -1
                        kern[bbp, bppbp] += fct_bppbp.real                      # kern[bbp, bppbp] += fct_bppbp.real
                        if bppbpi >= si.ndm0:
                            kern[bbp, bppbpi] -= fct_bppbp.imag*bppbp_sgn       # kern[bbp, bppbpi] -= fct_bppbp.imag*bppbp_sgn
                            if bbpi_bool:
                                kern[bbpi, bppbpi] += fct_bppbp.real*bppbp_sgn  # kern[bbpi, bppbpi] += fct_bppbp.real*bppbp_sgn
                        if bbpi_bool:
                            kern[bbpi, bppbp] += fct_bppbp.imag                 # kern[bbpi, bppbp] += fct_bppbp.imag
                    #--------------------------------------------------
                    bbpp = si.get_ind_dm0(b, bpp, charge)
                    if bbpp != -1:
                        fct_bbpp = 0
                        for a in si.statesdm[charge-1]:
                            for l in range(si.nleads):
                                fct_bbpp += -0.5*tLba[l, a, bpp].conjugate()*tLba[l, a, bp]
                        for c in si.statesdm[charge+1]:
                            for l in range(si.nleads):
                                fct_bbpp += -0.5*tLba[l, c, bpp].conjugate()*tLba[l, c, bp]
                        bbppi = si.ndm0 + bbpp - si.npauli
                        bbpp_sgn = +1 if si.get_ind_dm0(b, bpp, charge, maptype=3) else -1
                        kern[bbp, bbpp] += fct_bbpp.real                        # kern[bbp, bbpp] += fct_bbpp.real
                        if bbppi >= si.ndm0:
                            kern[bbp, bbppi] -= fct_bbpp.imag*bbpp_sgn          # kern[bbp, bbppi] -= fct_bbpp.imag*bbpp_sgn
                            if bbpi_bool:
                                kern[bbpi, bbppi] += fct_bbpp.real*bbpp_sgn     # kern[bbpi, bbppi] += fct_bbpp.real*bbpp_sgn
                        if bbpi_bool:
                            kern[bbpi, bbpp] += fct_bbpp.imag                   # kern[bbpi, bbpp] += fct_bbpp.imag
                #--------------------------------------------------
                for c, cp in itertools.product(si.statesdm[charge+1], si.statesdm[charge+1]):
                    ccp = si.get_ind_dm0(c, cp, charge+1)
                    if ccp != -1:
                        fct_ccp = 0
                        for l in range(si.nleads):
                            fct_ccp += tLba[l, b, c]*tLba[l, bp, cp].conjugate()
                        ccpi = si.ndm0 + ccp - si.npauli
                        ccp_sgn = +1 if si.get_ind_dm0(c, cp, charge+1, maptype=3) else -1
                        kern[bbp, ccp] += fct_ccp.real                          # kern[bbp, ccp] += fct_ccp.real
                        if ccpi >= si.ndm0:
                            kern[bbp, ccpi] -= fct_ccp.imag*ccp_sgn             # kern[bbp, ccpi] -= fct_ccp.imag*ccp_sgn
                            if bbpi_bool:
                                kern[bbpi, ccpi] += fct_ccp.real*ccp_sgn        # kern[bbpi, ccpi] += fct_ccp.real*ccp_sgn
                        if bbpi_bool:
                            kern[bbpi, ccp] += fct_ccp.imag                     # kern[bbpi, ccp] += fct_ccp.imag
                #--------------------------------------------------
    # Normalisation condition
    kern[norm_row] = np.zeros(si.ndm0r, dtype=doublenp)
    for charge in range(si.ncharge):
        for b in si.statesdm[charge]:
            bb = si.get_ind_dm0(b, b, charge)
            kern[norm_row, bb] += 1
    sys.kern = kern
    sys.bvec = bvec
    return 0

def generate_current_lindblad(sys):
    """
    Calculates currents using Lindblad approach.

    Parameters
    ----------
    sys : Approach
        Approach object.

    Modifies:
    sys.current : array
        Values of the current having nleads entries.
    sys.energy_current : array
        Values of the energy current having nleads entries.
    sys.heat_current : array
        Values of the heat current having nleads entries.
    """
    (phi0p, E, tLba, si) = (sys.phi0, sys.qd.Ea, sys.tLba, sys.si)
    current = np.zeros(si.nleads, dtype=complexnp)
    energy_current = np.zeros(si.nleads, dtype=complexnp)
    #
    phi0 = np.zeros(si.ndm0, dtype=complexnp)
    phi0[0:si.npauli] = phi0p[0:si.npauli]
    phi0[si.npauli:si.ndm0] = phi0p[si.npauli:si.ndm0] + 1j*phi0p[si.ndm0:]
    #
    for charge in range(si.ncharge):
        ccharge = charge+1
        bcharge = charge
        acharge = charge-1
        for b, bp in itertools.product(si.statesdm[bcharge], si.statesdm[bcharge]):
            bbp = si.get_ind_dm0(b, bp, bcharge)
            if bbp != -1:
                bbp_conj = si.get_ind_dm0(b, bp, bcharge, maptype=3)
                phi0bbp = phi0[bbp] if bbp_conj else phi0[bbp].conjugate()
                for l in range(si.nleads):
                    for a in si.statesdm[acharge]:
                        fcta = tLba[l, a, b]*phi0bbp*tLba[l, a, bp].conjugate()
                        current[l] -= fcta
                        energy_current[l] += (E[a]-0.5*(E[b]+E[bp]))*fcta
                    for c in si.statesdm[ccharge]:
                        fctc = tLba[l, c, b]*phi0bbp*tLba[l, c, bp].conjugate()
                        current[l] += fctc
                        energy_current[l] += (E[c]-0.5*(E[b]+E[bp]))*fctc
    sys.current = np.array(current.real, dtype=doublenp)
    sys.energy_current = np.array(energy_current.real, dtype=doublenp)
    sys.heat_current = sys.energy_current - sys.current*sys.leads.mulst
    return 0

def generate_vec_lindblad(phi0p, sys):
    """
    Acts on given phi0p with Liouvillian of Lindblad approach.

    Parameters
    ----------
    phi0p : array
        Some values of zeroth order density matrix elements.
    sys : Approach
        Approach object.

    Returns
    -------
    phi0 : array
        Values of zeroth order density matrix elements
        after acting with Liouvillian, i.e., phi0=L(phi0p).
    """
    (E, tLba, si, norm_row) = (sys.qd.Ea, sys.tLba, sys.si, sys.funcp.norm_row)
    #
    phi0 = np.zeros(si.ndm0, dtype=complexnp)
    phi0[0:si.npauli] = phi0p[0:si.npauli]
    phi0[si.npauli:si.ndm0] = phi0p[si.npauli:si.ndm0] + 1j*phi0p[si.ndm0:]
    #
    i_dphi0_dt = np.zeros(si.ndm0, dtype=complexnp)
    norm = 0
    for charge in range(si.ncharge):
        for b, bp in itertools.combinations_with_replacement(si.statesdm[charge], 2):
            bbp = si.get_ind_dm0(b, bp, charge)
            if bbp != -1:
                if b == bp: norm += phi0[bbp]
                bbp_bool = si.get_ind_dm0(b, bp, charge, maptype=2)
                if bbp_bool:
                    i_dphi0_dt[bbp] += (E[b]-E[bp])*phi0[bbp]
                    #--------------------------------------------------
                    for a, ap in itertools.product(si.statesdm[charge-1], si.statesdm[charge-1]):
                        aap = si.get_ind_dm0(a, ap, charge-1)
                        if aap != -1:
                            fct_aap = 0
                            for l in range(si.nleads):
                                fct_aap += tLba[l, b, a]*tLba[l, bp, ap].conjugate()
                            phi0aap = ( phi0[aap] if si.get_ind_dm0(a, ap, charge-1, maptype=3)
                                                  else phi0[aap].conjugate() )
                            i_dphi0_dt[bbp] += 1.0j*fct_aap*phi0aap
                    #--------------------------------------------------
                    for bpp in si.statesdm[charge]:
                        bppbp = si.get_ind_dm0(bpp, bp, charge)
                        if bppbp != -1:
                            fct_bppbp = 0
                            for a in si.statesdm[charge-1]:
                                for l in range(si.nleads):
                                    fct_bppbp += -0.5*tLba[l, a, b].conjugate()*tLba[l, a, bpp]
                            for c in si.statesdm[charge+1]:
                                for l in range(si.nleads):
                                    fct_bppbp += -0.5*tLba[l, c, b].conjugate()*tLba[l, c, bpp]
                            phi0bppbp = ( phi0[bppbp] if si.get_ind_dm0(bpp, bp, charge, maptype=3)
                                                      else phi0[bppbp].conjugate() )
                            i_dphi0_dt[bbp] += 1.0j*fct_bppbp*phi0bppbp
                        #--------------------------------------------------
                        bbpp = si.get_ind_dm0(b, bpp, charge)
                        if bbpp != -1:
                            fct_bbpp = 0
                            for a in si.statesdm[charge-1]:
                                for l in range(si.nleads):
                                    fct_bbpp += -0.5*tLba[l, a, bpp].conjugate()*tLba[l, a, bp]
                            for c in si.statesdm[charge+1]:
                                for l in range(si.nleads):
                                    fct_bbpp += -0.5*tLba[l, c, bpp].conjugate()*tLba[l, c, bp]
                            phi0bbpp = ( phi0[bbpp] if si.get_ind_dm0(b, bpp, charge, maptype=3)
                                                    else phi0[bbpp].conjugate() )
                            i_dphi0_dt[bbp] += 1.0j*fct_bbpp*phi0bbpp
                    #--------------------------------------------------
                    for c, cp in itertools.product(si.statesdm[charge+1], si.statesdm[charge+1]):
                        ccp = si.get_ind_dm0(c, cp, charge+1)
                        if ccp != -1:
                            fct_ccp = 0
                            for l in range(si.nleads):
                                fct_ccp += tLba[l, b, c]*tLba[l, bp, cp].conjugate()
                            phi0ccp = ( phi0[ccp] if si.get_ind_dm0(c, cp, charge+1, maptype=3)
                                                  else phi0[ccp].conjugate() )
                            i_dphi0_dt[bbp] += 1.0j*fct_ccp*phi0ccp
                    #--------------------------------------------------
    i_dphi0_dt[norm_row] = 1j*(norm-1)
    return np.concatenate((i_dphi0_dt.imag, i_dphi0_dt[si.npauli:si.ndm0].real))

class Approach_pyLindblad(Approach):

    kerntype = 'pyLindblad'
    generate_fct = staticmethod(generate_tLba)
    generate_kern = staticmethod(generate_kern_lindblad)
    generate_current = staticmethod(generate_current_lindblad)
    generate_vec = staticmethod(generate_vec_lindblad)
#---------------------------------------------------------------------------------------------------



"""Module containing cython functions, which solve 2vN approach integral equations.
   For docstrings see documentation of module neumann2."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import numpy as np
import itertools

from ..specfunc import kernel_fredriksen
from ..specfunc import hilbert_fredriksen
from .neumann2 import get_emin_emax
from .neumann2 import get_grid_ext
from .neumann2 import get_htransf_phi1k
from .neumann2 import get_htransf_fk
from .neumann2 import kern_phi0_2vN
from .neumann2 import generate_current_2vN

from ..mytypes import doublenp
from ..mytypes import complexnp

from ..aprclass import Approach2vN

cimport numpy as np
cimport cython

ctypedef np.uint8_t bool_t
ctypedef np.int_t int_t
ctypedef np.int64_t long_t
ctypedef np.float64_t double_t
ctypedef np.complex128_t complex_t

#from scipy import pi as scipy_pi
#cdef double_t pi = scipy_pi
cdef double_t pi = 3.14159265358979323846

@cython.cdivision(True)
@cython.boundscheck(False)
cdef complex_t func_2vN(double_t Ek,
                        np.ndarray[double_t, ndim=1] Ek_grid,
                        int_t l,
                        int_t eta,
                        np.ndarray[complex_t, ndim=2] hfk):
    cdef long_t b_idx, a_idx
    cdef double_t a, b
    cdef complex_t fa, fb, rez
    if Ek<Ek_grid[0] or Ek>Ek_grid[-1]:
        return 0
    #
    b_idx = (<long_t>((Ek-Ek_grid[0])/(Ek_grid[1]-Ek_grid[0])))+1
    #if b_idx == len(Ek_grid): b_idx -= 1
    if b_idx == Ek_grid.shape[0]: b_idx -= 1
    a_idx = b_idx - 1
    b, a = Ek_grid[b_idx], Ek_grid[a_idx]
    #
    fb = hfk[l, b_idx]
    fa = hfk[l, a_idx]
    rez = Ek/(b-a)*(fb-fa) + 1/(b-a)*(b*fa-a*fb)
    return pi*rez if eta+1 else pi*rez.conjugate()

@cython.cdivision(True)
@cython.boundscheck(False)
cdef int_t get_at_k1(double_t Ek,
                     np.ndarray[double_t, ndim=1] Ek_grid,
                     int_t l,
                     long_t cb,
                     bint conj,
                     np.ndarray[complex_t, ndim=4] phi1k,
                     np.ndarray[complex_t, ndim=4] hphi1k,
                     long_t ndm0,
                     complex_t fct,
                     int_t eta,
                     np.ndarray[complex_t, ndim=1] term):
    #
    cdef long_t b_idx, a_idx, bbp
    cdef double_t a, b
    cdef complex_t fa, fb, u, hu
    if Ek<Ek_grid[0] or Ek>Ek_grid[-1]:
        return 0
    #
    b_idx = (<long_t>((Ek-Ek_grid[0])/(Ek_grid[1]-Ek_grid[0])))+1
    #NOTE This line needs to be optimized
    #if b_idx == len(Ek_grid): b_idx -= 1
    if b_idx == Ek_grid.shape[0]: b_idx -= 1
    a_idx = b_idx - 1
    b, a = Ek_grid[b_idx], Ek_grid[a_idx]
    #
    for bbp in range(ndm0):
        #fa = phi1k[a_idx, l, cb, bbp].conjugate() if conj else phi1k[a_idx, l, cb, bbp]
        #fb = phi1k[b_idx, l, cb, bbp].conjugate() if conj else phi1k[b_idx, l, cb, bbp]
        #u = Ek/(b-a)*(fb-fa) + 1/(b-a)*(b*fa-a*fb)
        fa = phi1k[a_idx, l, cb, bbp]
        fb = phi1k[b_idx, l, cb, bbp]
        u = Ek/(b-a)*(fb-fa) + 1/(b-a)*(b*fa-a*fb)
        u = u.conjugate() if conj else u
        #
        #fa = hphi1k[a_idx, l, cb, bbp].conjugate() if conj else hphi1k[a_idx, l, cb, bbp]
        #fb = hphi1k[b_idx, l, cb, bbp].conjugate() if conj else hphi1k[b_idx, l, cb, bbp]
        #hu = Ek/(b-a)*(fb-fa) + 1/(b-a)*(b*fa-a*fb)
        fa = hphi1k[a_idx, l, cb, bbp]
        fb = hphi1k[b_idx, l, cb, bbp]
        hu = Ek/(b-a)*(fb-fa) + 1/(b-a)*(b*fa-a*fb)
        hu = hu.conjugate() if conj else hu
        #
        term[bbp] = term[bbp] + pi*fct*(hu+eta*1j*u)
    return 0

@cython.cdivision(True)
@cython.boundscheck(False)
def c_phi1k_local_2vN(long_t ind,
                      np.ndarray[double_t, ndim=1] Ek_grid,
                      np.ndarray[double_t, ndim=2] fk,
                      np.ndarray[complex_t, ndim=2] hfkp,
                      np.ndarray[complex_t, ndim=2] hfkm,
                      np.ndarray[double_t, ndim=1] E,
                      np.ndarray[complex_t, ndim=3] Tba,
                      si):
    cdef int_t acharge, bcharge, ccharge, dcharge, charge, l, l1, nleads, itype, dqawc_limit
    cdef long_t c, b, cb, a1, b1, c1, d1, b1a1, b1b, cb1, c1b, cc1, d1c1
    cdef double_t fp, fm
    cdef double_t Ek = Ek_grid[ind]
    nleads = si.nleads
    #
    cdef np.ndarray[long_t, ndim=1] lenlst = si.lenlst
    cdef np.ndarray[long_t, ndim=1] dictdm = si.dictdm
    cdef np.ndarray[long_t, ndim=1] shiftlst0 = si.shiftlst0
    cdef np.ndarray[long_t, ndim=1] shiftlst1 = si.shiftlst1
    cdef np.ndarray[long_t, ndim=1] mapdm0 = si.mapdm0
    #
    cdef np.ndarray[complex_t, ndim=3] kern0 = np.zeros((nleads, si.ndm1, si.ndm0), dtype=complexnp)
    cdef np.ndarray[complex_t, ndim=3] kern1 = np.zeros((nleads, si.ndm1, si.ndm1), dtype=complexnp)
    for charge in range(si.ncharge-1):
        dcharge = charge+2
        ccharge = charge+1
        bcharge = charge
        acharge = charge-1
        for c, b in itertools.product(si.statesdm[ccharge], si.statesdm[bcharge]):
            cb = lenlst[bcharge]*dictdm[c] + dictdm[b] + shiftlst1[bcharge]
            for l in range(nleads):
                # Note that the bias is put in the distributions and not in the dispersion
                kern1[l, cb, cb] = kern1[l, cb, cb] + (Ek-E[c]+E[b]+0j)
                fp = fk[l, ind]  #fermi_func(+(Ek-mulst[l])/tlst[l])
                fm = 1.-fp       #fermi_func(-(Ek-mulst[l])/tlst[l])
                # Phi[0] terms
                for b1 in si.statesdm[bcharge]:
                    b1b = mapdm0[lenlst[bcharge]*dictdm[b1] + dictdm[b] + shiftlst0[bcharge]]
                    kern0[l, cb, b1b] = kern0[l, cb, b1b] + fp*Tba[l, c, b1]
                for c1 in si.statesdm[ccharge]:
                    cc1 = mapdm0[lenlst[ccharge]*dictdm[c] + dictdm[c1] + shiftlst0[ccharge]]
                    kern0[l, cb, cc1] = kern0[l, cb, cc1] - fm*Tba[l, c1, b]
                #---------------------------------------------------------------------------
                # Phi[1] terms
                # 2nd and 7th terms
                for b1, a1 in itertools.product(si.statesdm[bcharge], si.statesdm[acharge]):
                    b1a1 = lenlst[acharge]*dictdm[b1] + dictdm[a1] + shiftlst1[acharge]
                    for l1 in range(nleads):
                        kern1[l, cb, b1a1] = kern1[l, cb, b1a1] - Tba[l1, c, b1]*Tba[l1, a1, b]*(
                                              +func_2vN(+(Ek-E[b1]+E[b]), Ek_grid, l1, +1, hfkp)
                                              -func_2vN(-(Ek-E[c]+E[a1]), Ek_grid, l1, -1, hfkp) )
                # 6th and 8th terms
                for b1 in si.statesdm[bcharge]:
                    cb1 = lenlst[bcharge]*dictdm[c] + dictdm[b1] + shiftlst1[bcharge]
                    for l1 in range(nleads):
                        for c1 in si.statesdm[ccharge]:
                            kern1[l, cb, cb1] = kern1[l, cb, cb1] - (Tba[l1, b1, c1]*Tba[l1, c1, b]
                                                 *func_2vN(+(Ek-E[c]+E[c1]), Ek_grid, l1, +1, hfkp))
                        for a1 in si.statesdm[acharge]:
                            kern1[l, cb, cb1] = kern1[l, cb, cb1] + (Tba[l1, b1, a1]*Tba[l1, a1, b]
                                                 *func_2vN(-(Ek-E[c]+E[a1]), Ek_grid, l1, -1, hfkm))
                # 1st and 3rd terms
                for c1 in si.statesdm[ccharge]:
                    c1b = lenlst[bcharge]*dictdm[c1] + dictdm[b] + shiftlst1[bcharge]
                    for l1 in range(nleads):
                        for b1 in si.statesdm[bcharge]:
                            kern1[l, cb, c1b] = kern1[l, cb, c1b] - (Tba[l1, c, b1]*Tba[l1, b1, c1]
                                                 *func_2vN(+(Ek-E[b1]+E[b]), Ek_grid, l1, +1, hfkm))
                        for d1 in si.statesdm[dcharge]:
                            kern1[l, cb, c1b] = kern1[l, cb, c1b] + (Tba[l1, c, d1]*Tba[l1, d1, c1]
                                                 *func_2vN(-(Ek-E[d1]+E[b]), Ek_grid, l1, -1, hfkp))
                #
                for d1, c1 in itertools.product(si.statesdm[dcharge], si.statesdm[ccharge]):
                    d1c1 = lenlst[ccharge]*dictdm[d1] + dictdm[c1] + shiftlst1[ccharge]
                    for l1 in range(nleads):
                        kern1[l, cb, d1c1] = kern1[l, cb, d1c1] - Tba[l1, c, d1]*Tba[l1, c1, b]*(
                                              +func_2vN(+(Ek-E[c]+E[c1]), Ek_grid, l1, +1, hfkm)
                                              -func_2vN(-(Ek-E[d1]+E[b]), Ek_grid, l1, -1, hfkm) )
    for l in range(nleads):
        kern0[l] = np.dot(np.linalg.inv(kern1[l]), kern0[l])
    return kern0, kern1

@cython.cdivision(True)
@cython.boundscheck(False)
def c_phi1k_iterate_2vN(long_t ind,
                        np.ndarray[double_t, ndim=1] Ek_grid,
                        np.ndarray[complex_t, ndim=4] phi1k,
                        np.ndarray[complex_t, ndim=4] hphi1k,
                        np.ndarray[double_t, ndim=2] fk,
                        np.ndarray[complex_t, ndim=3] kern1,
                        np.ndarray[double_t, ndim=1] E,
                        np.ndarray[complex_t, ndim=3] Tba,
                        si):
    #
    cdef int_t acharge, bcharge, ccharge, dcharge, charge, l, l1, nleads
    cdef long_t ndm0, c, b, cb, a1, b1, c1, d1, ba1, c1b1, c1b, d1c1, d1c, b1a1, cb1, bbp
    cdef double_t fp, fm
    cdef double_t Ek = Ek_grid[ind]
    ndm0 = si.ndm0
    nleads = si.nleads
    #
    cdef np.ndarray[long_t, ndim=1] lenlst = si.lenlst
    cdef np.ndarray[long_t, ndim=1] dictdm = si.dictdm
    cdef np.ndarray[long_t, ndim=1] shiftlst0 = si.shiftlst0
    cdef np.ndarray[long_t, ndim=1] shiftlst1 = si.shiftlst1
    cdef np.ndarray[long_t, ndim=1] mapdm0 = si.mapdm0
    #
    cdef np.ndarray[complex_t, ndim=3] kern0 = np.zeros((si.nleads, si.ndm1, si.ndm0), dtype=complexnp)
    cdef np.ndarray[complex_t, ndim=1] term = np.zeros(ndm0, dtype=complexnp)
    #
    for charge in range(si.ncharge-1):
        dcharge = charge+2
        ccharge = charge+1
        bcharge = charge
        acharge = charge-1
        for c, b in itertools.product(si.statesdm[ccharge], si.statesdm[bcharge]):
            cb = lenlst[bcharge]*dictdm[c] + dictdm[b] + shiftlst1[bcharge]
            for l in range(nleads):
                fp = fk[l, ind] #fermi_func(Ek/tlst[l])
                fm = 1.-fp
                # 1st term
                for a1 in si.statesdm[acharge]:
                    ba1 = lenlst[acharge]*dictdm[b] + dictdm[a1] + shiftlst1[acharge]
                    for b1, l1 in itertools.product(si.statesdm[bcharge], range(nleads)):
                        #print('1')
                        get_at_k1(+(Ek-E[b1]+E[b]), Ek_grid, l1, ba1, True, phi1k, hphi1k,
                                  ndm0, -fp*Tba[l1, c, b1]*Tba[l, b1, a1], -1, term) #*
                # 2nd and 5th terms
                for b1, c1 in itertools.product(si.statesdm[bcharge], si.statesdm[ccharge]):
                    #print('2 and 5')
                    c1b1 = lenlst[bcharge]*dictdm[c1] + dictdm[b1] + shiftlst1[bcharge]
                    for l1 in range(nleads):
                        # 2nd term
                        get_at_k1(+(Ek-E[b1]+E[b]), Ek_grid, l1, c1b1, True, phi1k, hphi1k,
                                  ndm0, -fm*Tba[l1, c, b1]*Tba[l, c1, b], -1, term)
                        # 5th term
                        get_at_k1(+(Ek-E[c]+E[c1]), Ek_grid, l1, c1b1, True, phi1k, hphi1k,
                                  ndm0, -fp*Tba[l, c, b1]*Tba[l1, c1, b], -1, term)
                # 3rd term
                for c1 in si.statesdm[ccharge]:
                    c1b = lenlst[bcharge]*dictdm[c1] + dictdm[b] + shiftlst1[bcharge]
                    for d1, l1 in itertools.product(si.statesdm[dcharge], range(nleads)):
                        #print('3')
                        get_at_k1(-(Ek-E[d1]+E[b]), Ek_grid, l1, c1b, False, phi1k, hphi1k,
                                  ndm0, +fp*Tba[l1, c, d1]*Tba[l, d1, c1], +1, term)
                # 4th term
                for d1, c1 in itertools.product(si.statesdm[dcharge], si.statesdm[ccharge]):
                    #print('4')
                    d1c1 = lenlst[ccharge]*dictdm[d1] + dictdm[c1] + shiftlst1[ccharge]
                    for l1 in range(nleads):
                        get_at_k1(-(Ek-E[d1]+E[b]), Ek_grid, l1, d1c1, False, phi1k, hphi1k,
                                  ndm0, +fm*Tba[l1, c, d1]*Tba[l, c1, b], +1, term)
                # 6th term
                for d1 in si.statesdm[dcharge]:
                    d1c = lenlst[ccharge]*dictdm[d1] + dictdm[c] + shiftlst1[ccharge]
                    for c1, l1 in itertools.product(si.statesdm[ccharge], range(nleads)):
                        #print('6')
                        get_at_k1(+(Ek-E[c]+E[c1]), Ek_grid, l1, d1c, True, phi1k, hphi1k,
                                  ndm0, -fm*Tba[l, d1, c1]*Tba[l1, c1, b], -1, term)
                # 7th term
                for b1, a1 in itertools.product(si.statesdm[bcharge], si.statesdm[acharge]):
                    #print('7')
                    b1a1 = lenlst[acharge]*dictdm[b1] + dictdm[a1] + shiftlst1[acharge]
                    for l1 in range(nleads):
                        get_at_k1(-(Ek-E[c]+E[a1]), Ek_grid, l1, b1a1, False, phi1k, hphi1k,
                                  ndm0, +fp*Tba[l, c, b1]*Tba[l1, a1, b], +1, term)
                # 8th term
                for b1 in si.statesdm[bcharge]:
                    cb1 = lenlst[bcharge]*dictdm[c] + dictdm[b1] + shiftlst1[bcharge]
                    for a1, l1 in itertools.product(si.statesdm[acharge], range(nleads)):
                        #print('8')
                        get_at_k1(-(Ek-E[c]+E[a1]), Ek_grid, l1, cb1, False, phi1k, hphi1k,
                                  ndm0, +fm*Tba[l, b1, a1]*Tba[l1, a1, b], +1, term)
                for bbp in range(ndm0):
                    kern0[l, cb, bbp] = term[bbp]
                    term[bbp] = 0
    for l in range(nleads):
        kern0[l] = np.dot(np.linalg.inv(kern1[l]), kern0[l])
    return kern0

@cython.boundscheck(False)
def c_get_phi1_phi0_2vN(sys):
    cdef long_t j1, Eklen
    cdef double_t dx
    cdef np.ndarray[double_t, ndim=1] Ek_grid = sys.Ek_grid
    #
    (phi1k, si) = (sys.phi1k, sys.si)
    # Get integrated Phi[1]_{cb} in terms of Phi[0]_{bb'}
    #cdef np.ndarray[complex_t, ndim=3]
    phi1_phi0 = np.zeros((si.nleads, si.ndm1, si.ndm0), dtype=complexnp)
    e_phi1_phi0 = np.zeros((si.nleads, si.ndm1, si.ndm0), dtype=complexnp)
    Eklen = Ek_grid.shape[0] #len(Ek_grid)
    #
    for j1 in range(Eklen):
        # Trapezoidal rule
        if j1 == 0:         dx = Ek_grid[j1+1] - Ek_grid[j1]
        elif j1 == Eklen-1: dx = Ek_grid[j1]   - Ek_grid[j1-1]
        else:               dx = Ek_grid[j1+1] - Ek_grid[j1-1]
        phi1_phi0 += 0.5*dx*phi1k[j1]
        e_phi1_phi0 += 0.5*dx*Ek_grid[j1]*phi1k[j1]
    sys.phi1_phi0 = phi1_phi0
    sys.e_phi1_phi0 = e_phi1_phi0
    return 0

@cython.boundscheck(False)
def c_iterate_2vN(sys):
    cdef long_t j1, Eklen, ind, kpnt_left
    cdef np.ndarray[double_t, ndim=1] Ek_grid = sys.Ek_grid
    #cdef np.ndarray[double_t, ndim=1] Ek_grid_ext = sys.Ek_grid_ext
    cdef np.ndarray[double_t, ndim=1] E = sys.qd.Ea
    cdef np.ndarray[complex_t, ndim=3] Tba = sys.leads.Tba
    #
    (Ek_grid_ext) = (sys.Ek_grid_ext)
    (si, funcp) = (sys.si, sys.funcp)
    (mulst, tlst) = (sys.leads.mulst, sys.leads.tlst)
    # Assign sys.phi1k_delta to phi1k_delta_old, because the new phi1k_delta
    # will be generated in this function
    (phi1k_delta_old, kern1k) = (sys.phi1k_delta, sys.kern1k)
    #
    Eklen = Ek_grid.shape[0] #len(Ek_grid)
    if phi1k_delta_old is None:
        ## Define the extended grid Ek_grid_ext for calculations outside the bandwidth
        # Here sys.funcp.emin, sys.funcp.emax are defined
        get_emin_emax(sys)
        # Here sys.Ek_grid_ext, sys.funcp.kpnt_left, sys.funcp.kpnt_right are defined
        get_grid_ext(sys)
        Ek_grid_ext = sys.Ek_grid_ext
        Eklen_ext = Ek_grid_ext.shape[0] #len(Ek_grid_ext)
        # Generate the Fermi functions on the grid
        # This is necessary to generate only if Ek_grid, mulst, or tlst are changed
        sys.fkp = np.zeros((si.nleads, Eklen), dtype=doublenp)
        for l in range(si.nleads):
            sys.fkp[l] = 1/( np.exp((Ek_grid - mulst[l])/tlst[l]) + 1 )
        sys.fkm = 1-sys.fkp
        sys.fkp, sys.hfkp = get_htransf_fk(sys.fkp, funcp)
        sys.fkm, sys.hfkm = get_htransf_fk(sys.fkm, funcp)
        # Calculate the zeroth iteration of Phi[1](k)
        phi1k_delta = np.zeros((Eklen, si.nleads, si.ndm1, si.ndm0), dtype=complexnp)
        kern1k = np.zeros((Eklen, si.nleads, si.ndm1, si.ndm1), dtype=complexnp)
        kpnt_left = funcp.kpnt_left
        for j1 in range(Eklen):
            ind = j1 + kpnt_left
            phi1k_delta[j1], kern1k[j1] = c_phi1k_local_2vN(ind, Ek_grid_ext, sys.fkp,
                                                            sys.hfkp, sys.hfkm, E, Tba, si)
        hphi1k_delta = None
    elif kern1k is None:
        pass
    else:
        # Hilbert transform phi1k_delta_old on extended grid Ek_grid_ext
        #print('Hilbert transforming')
        phi1k_delta_old, hphi1k_delta = get_htransf_phi1k(phi1k_delta_old, funcp)
        #print('Making an iteration')
        phi1k_delta = np.zeros((Eklen, si.nleads, si.ndm1, si.ndm0), dtype=complexnp)
        kpnt_left = funcp.kpnt_left
        for j1 in range(Eklen):
            ind = j1 + kpnt_left
            phi1k_delta[j1] = c_phi1k_iterate_2vN(ind, Ek_grid_ext, phi1k_delta_old, hphi1k_delta,
                                                  sys.fkp, kern1k[j1], E, Tba, si)
    sys.phi1k_delta = phi1k_delta
    sys.hphi1k_delta = hphi1k_delta
    sys.kern1k = kern1k
    return 0

class Approach_2vN(Approach2vN):

    kerntype = '2vN'
    iterate = c_iterate_2vN
    get_phi1_phi0 = c_get_phi1_phi0_2vN
    kern_phi0 = staticmethod(kern_phi0_2vN)
    generate_current = staticmethod(generate_current_2vN)



"""
QmeQ: Quantum master equation for Quantum dot transport calculations
====================================================================

QmeQ is an open-source Python package for transport calculations through
quantum  dot devices. The so-called Anderson-type models are used to describe
the quantum dot device, where quantum dots are coupled to the leads by
tunneling. QmeQ can calculate the stationary state particle and energy currents
using various approximate density matrix approaches. As for now we have
implemented the following first-order methods

* Pauli (classical) master equation
* Lindblad approach
* Redfield approach
* First order von Neumann (1vN) approach

which can describe the effect of Coulomb blockade. QmeQ also has one
second-order method

* Second order von Neumann (2vN) approach

which can additionally address cotunneling, pair tunneling, and
broadening effects.

Physics disclaimer
------------------

All the methods in QmeQ are approximate so depending on parameter regime they
can fail, and a good knowledge of the method is required whether to trust the
result or not. For example, Redfield, 1vN, and 2vN approaches can violate
positivity of the reduced density matrix and lead to currents flowing against
the bias. We still think it is important to have a package where a user can
duplicate existing calculations, check applicability of different methods, or
simply discover new kind of physics using different approximate master equations.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from .aprclass import Approach
from .aprclass import Approach2vN
from .builder import Builder
from .builder import FunctionProperties
from .indexing import StateIndexing
from .indexing import StateIndexingPauli
from .indexing import StateIndexingDM
from .indexing import StateIndexingDMc
from .leadstun import LeadsTunneling
from .qdot import QuantumDot

__version__ = '1.0'



"""Module containing various special functions."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np
from numpy.fft import fft, ifft
from numpy import sign
from scipy import pi
from scipy import log
from scipy import log2
from scipy import exp
from scipy import sqrt
from scipy.special import psi as digamma
from scipy.integrate import quad

from .mytypes import complexnp
from .mytypes import doublenp
from .mytypes import intnp

def fermi_func(x):
    """Fermi function."""
    return 1/(exp(x)+1)

def func_pauli(Ecb, mu, T, Dm, Dp, itype):
    """
    Function used when generating Pauli master equation kernel.

    Parameters
    ----------
    Ecb : float
        Energy.
    mu : float
        Chemical potential.
    T : float
        Temperature.
    Dm,Dp : float
        Bandwidth.

    Returns
    -------
    array
        | Array of two float numbers [cur0, cur1] containing
          momentum-integrated current amplitudes.
        | cur0 - particle current amplitude.
        | cur1 - hole current amplitude.
    """
    alpha = (Ecb-mu)/T
    Rm, Rp = (Dm-mu)/T, (Dp-mu)/T
    if itype == 1 or itype == 3 or (alpha < Rp and alpha > Rm):
        cur0 = fermi_func(alpha)
        cur1 = 1-cur0
        rez = 2*pi*np.array([cur0, cur1])
    else:
        rez = np.zeros(2)
    return rez

def func_1vN(Ecb, mu, T, Dm, Dp, itype, limit):
    """
    Function used when generating 1vN, Redfield approach kernel.

    Parameters
    ----------
    Ecb : float
        Energy.
    mu : float
        Chemical potential.
    T : float
        Temperature.
    Dm,Dp : float
        Bandwidth.
    itype : int
        | Type of integral for first order approach calculations.
        | itype=0: the principal parts are evaluated using Fortran integration package QUADPACK \
                   routine dqawc through SciPy.
        | itype=1: the principal parts are kept, but approximated by digamma function valid for \
                   large bandwidht D.
        | itype=2: the principal parts are neglected.
        | itype=3: the principal parts are neglected and infinite bandwidth D is assumed.
    limit : int
        For itype=0 dqawc_limit determines the maximum number of subintervals
        in the partition of the given integration interval.

    Returns
    -------
    array
        | Array of four complex numbers [cur0, cur1, en0, en1] containing
          momentum-integrated current amplitudes.
        | cur0 - particle current amplitude.
        | cur1 - hole current amplitude.
        | en0 - particle energy current amplitude.
        | en1 - hol energy current amplitude.
    """
    if itype == 0:
        alpha, Rm, Rp = (Ecb-mu)/T, (Dm-mu)/T, (Dp-mu)/T
        cur0, err = quad(fermi_func, Rm, Rp, weight='cauchy', wvar=alpha, epsabs=1.0e-6,
                                             epsrel=1.0e-6, limit=limit)
        cur0 = cur0 + (-1.0j*pi*fermi_func(alpha) if alpha < Rp and alpha > Rm else 0)
        cur1 = cur0 + log(abs((Rm-alpha)/(Rp-alpha)))
        cur1 = cur1 + (1.0j*pi if alpha < Rp and alpha > Rm else 0)
        #
        const0 = T*((-Rm if Rm < -40 else log(1+exp(-Rm)))
                   -(-Rp if Rp < -40 else log(1+exp(-Rp))))
        const1 = const0 + Dm-Dp
        #
        en0 = const0 + Ecb*cur0
        en1 = const1 + Ecb*cur1
    elif itype == 1:
        alpha, Rm, Rp = (Ecb-mu)/T, Dm/T, Dp/T
        cur0 = digamma(0.5+1.0j*alpha/(2*pi)).real - log(abs(Rm)/(2*pi))
        cur0 = cur0 - 1.0j*pi*fermi_func(alpha)
        cur1 = cur0 + log(abs(Rm/Rp))
        cur1 = cur1 + 1.0j*pi
        #
        en0 = -T*Rm + Ecb*cur0
        en1 = -T*Rp + Ecb*cur1
    elif itype == 2:
        alpha, Rm, Rp = (Ecb-mu)/T, (Dm-mu)/T, (Dp-mu)/T
        cur0 = -1.0j*pi*fermi_func(alpha) if alpha < Rp and alpha > Rm else 0
        cur1 = cur0 + (1.0j*pi if alpha < Rp and alpha > Rm else 0)
        en0 = Ecb*cur0
        en1 = Ecb*cur1
    elif itype == 3:
        alpha = (Ecb-mu)/T
        cur0 = -1.0j*pi*fermi_func(alpha)
        cur1 = cur0 + 1.0j*pi
        en0 = Ecb*cur0
        en1 = Ecb*cur1
    #-------------------------
    return np.array([cur0, cur1, en0, en1])

def kernel_fredriksen(n):
    """
    Generates kernel for Hilbert transform using FFT.

    Parameters
    ----------
    n : int
        Number of equidistant grid points.

    Returns
    -------
    array
        Kernel used when performing Hilbert transform using FFT.
    """
    aux = np.zeros(n+1, dtype=doublenp)
    for i in range(1,n+1):
        aux[i] = i*log(i)
    m = 2*n
    ker = np.zeros(m, dtype=doublenp)
    for i in range(1,n):
        ker[i] = aux[i+1]-2*aux[i]+aux[i-1]
        ker[m-i] = -ker[i]
    return fft(ker)/pi

def hilbert_fredriksen(f, ker=None):
    """
    Performs Hilbert transform of f.

    Parameters
    ----------
    f : array
        Values of function on a equidistant grid.
    ker : array
        Kernel used when performing Hilbert transform using FFT.

    Returns
    -------
    array
        Hilbert transform of f.
    """
    if ker is None:
        ker = kernel_fredriksen(len(f))
    n = len(f)
    fpad = fft(np.concatenate( (f,np.zeros(len(ker)-n)) ))
    r = ifft(fpad*ker)
    return r[0:n]



"""Module containing sample classes for approach definitions."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np
import scipy.sparse.linalg
from scipy import optimize

from .mytypes import doublenp
from .mytypes import complexnp

class Approach(object):
    """
    Sample class used to define different approaches.

    Attributes
    ----------
    qd : QuantumDot
        QuantumDot object.
    leads : LeadsTunneling
        LeadsTunneling object.
    si : StateIndexingDM
        StateIndexingDM object.
    kern : array
        Kernel (Liouvillian) representing the master equation.
    bvec : array
        Right hand side column vector for master equation.
        The entry funcp.norm_row is 1 representing normalization condition.
    sol0 : array
        Least squares solution for the master equation.
    phi0 : array
        Values of zeroth order density matrix elements.
    phi1 : array
        Values of first order density matrix elements
        stored in nleads by ndm1 numpy array.
    current : array
        Values of the current having nleads entries.
    energy_current : array
        Values of the energy current having nleads entries.
    heat_current : array
        Values of the heat current having nleads entries.
    funcp : FunctionProperties
        FunctionProperties object.
    paulifct : array
        Factors used for generating Pauli master equation kernel.
    phi1fct : array
        Factors used for generating 1vN, Redfield master equation kernels.
    phi1fct_energy : array
        Factors used to calculate energy and heat currents in 1vN, Redfield approaches.
    tLba : array
        Jump operator matrix in many-body basis for Lindblad approach.
    """

    kerntype = 'not defined'
    indexing_class_name = 'StateIndexingDM'

    @staticmethod
    def generate_fct(sys): pass
    @staticmethod
    def generate_kern(sys): pass
    @staticmethod
    def generate_current(sys): pass
    @staticmethod
    def generate_vec(sys): pass

    def __init__(self, sys):
        """
        Initialization of the Approach class.

        Parameters
        ----------
        sys : Approach, Builder, etc. object.
            Any object having qd, leads, si, and funcp attributes.
        """
        self.qd = sys.qd
        self.leads = sys.leads
        self.si = sys.si
        self.funcp = sys.funcp
        self.restart()

    def restart(self):
        """Restart values of some variables."""
        self.kern, self.bvec = None, None
        self.sol0, self.phi0, self.phi1 = None, None, None
        self.current = None
        self.energy_current = None
        self.heat_current = None
        self.phi1fct, self.paulifct = None, None

    def set_phi0_init(self):
        if self.kerntype in {'Pauli', 'pyPauli'}:
            phi0_init = np.zeros(self.si.npauli, dtype=doublenp)
        else:
            phi0_init = np.zeros(self.si.ndm0r, dtype=doublenp)
        phi0_init[0] = 1.0
        return phi0_init

    def solve_kern(self):
        """Finds the stationary state using least squares or by inverting the matrix."""
        solmethod = self.funcp.solmethod
        symq = self.funcp.symq
        if solmethod == 'n': solmethod = 'solve' if symq else 'lsqr'
        if not symq and solmethod != 'lsqr' and solmethod != 'lsmr':
            print("WARNING: Using solmethod=lsqr, because the kernel is not symmetric, symq=False.")
            solmethod = 'lsqr'
        try:
            if   solmethod == 'solve': self.sol0 = [np.linalg.solve(self.kern, self.bvec)]
            elif solmethod == 'lsqr':  self.sol0 = np.linalg.lstsq(self.kern, self.bvec)
            self.phi0 = self.sol0[0]
        except Exception as exept:
            self.funcp.print_error(exept)
            self.phi0 = np.zeros(self.si.ndm0r)
        self.funcp.solmethod = solmethod

    def solve_matrix_free(self):
        """Finds the stationary state using matrix free methods like broyden, krylov, etc."""
        solmethod = self.funcp.solmethod
        #
        phi0_init = self.funcp.phi0_init
        if phi0_init is None:
            self.funcp.print_warning(0, "WARNING: For mfreeq=True no phi0_init is specified. "+
                                        "Using phi0_init[0]=1.0 as a default. "+
                                        "This warning will not be shown again.")
            phi0_init = self.set_phi0_init()
        #
        solmethod = solmethod if solmethod != 'n' else 'krylov'
        try:
            self.sol0 = optimize.root(self.generate_vec, phi0_init, args=(self), method=solmethod)
            self.phi0 = self.sol0.x
        except Exception as exept:
            self.funcp.print_error(exept)
            self.phi0 = 0*phi0_init
        self.funcp.solmethod = solmethod

    def solve(self, qdq=True, rotateq=True, masterq=True, currentq=True, *args, **kwargs):
        """
        Solves the master equation.

        Parameters
        ----------
        qdq : bool
            Diagonalise many-body quantum dot Hamiltonian
            and express the lead matrix Tba in the eigenbasis.
        rotateq : bool
            Rotate the many-body tunneling matrix Tba.
        masterq : bool
            Solve the master equation.
        currentq : bool
            Calculate the current.
        """
        if qdq:
            self.qd.diagonalise()
            if rotateq:
                self.leads.rotate(self.qd.vecslst)
        #
        if masterq:
            self.generate_fct(self)
            if self.funcp.mfreeq:
                self.solve_matrix_free()
            else:
                self.generate_kern(self)
                self.solve_kern()
            if currentq:
                self.generate_current(self)


class Iterations2vN(object):
    """
    Class for storing some properties of the system after 2vN iteration.
    """

    def __init__(self, sys):
        self.niter = sys.niter
        self.phi0 = sys.phi0
        self.phi1 = sys.phi1
        self.current = sys.current
        self.energy_current = sys.energy_current
        self.heat_current = sys.heat_current


class Approach2vN(Approach):
    """
    Sample class for solving the 2vN approach for stationary state.
    Most of the attributes are the same as in Approach class.
    Here we define the Python implementation of the 2vN approach.

    Attributes
    ----------
    kern : array
        Kernel for zeroth order density matrix elements Phi[0].
    funcp.kpnt : int
        Number of energy grid points on which 2vN approach equations are solved.
    Ek_grid : array
        Energy grid on which 2vN approach equations are solved.
    Ek_grid_ext : array
        Extension of Ek_grid by neumann2py.get_grid_ext(sys).
    niter : int
        Number of iterations performed when solving integral equation.
    iters : Iterations2vN
        Iterations2vN object.
    phi1k : array
        Numpy array with dimensions (len(Ek_grid), nleads, ndm1, ndm0)
        cotaining energy resolved first order density matrix elements Phi[1](k).
    phi1k_delta : array
        Numpy array with dimensions (len(Ek_grid), nleads, ndm1, ndm0)
        Difference from phi1k after performing one iteration.
    hphi1k_delta : array
        Numpy array with dimensions (len(Ek_grid_ext), nleads, ndm1, ndm0)
        Hilbert transform of phi1k_delta.
    kern1k : array
        Numpy array with dimensions (len(Ek_grid), nleads, ndm1, ndm1)
        corresponding to energy resolved local kernel for Phi[1](k).
    fkp, fkm : array
        nleads by len(Ek_grid_ext) numpy array containing
        Fermi function (fkp) and 1-Fermi (fkm) values on the grid Ek_grid_ext.
    hfkp, hfkm : array
        Hilbert transform of fkp, fkm.
    """

    kerntype = 'not defined'
    indexing_class_name = 'StateIndexingDMc'
    @staticmethod
    def iterate(sys): pass
    @staticmethod
    def get_phi1_phi0(sys): pass
    @staticmethod
    def kern_phi0(sys): pass
    @staticmethod
    def generate_current(sys): pass

    def __init__(self, sys):
        """
        Initialization of the Approach2vN class.

        Parameters
        ----------
        sys : Approach, Builder, etc. object.
            Any object having qd, leads, si, and funcp attributes.
        """
        self.qd = sys.qd
        self.leads = sys.leads
        self.si = sys.si
        self.funcp = sys.funcp
        self.fkp, self.fkm = None, None
        self.hfkp, self.hfkm = None, None
        self.restart()
        self.Ek_grid = np.zeros(1, dtype=doublenp)
        self.Ek_grid_ext = np.zeros(0, dtype=doublenp)
        # Some exceptions
        if type(self.si).__name__ != self.indexing_class_name:
            raise TypeError('The state indexing class for 2vN approach has to be StateIndexingDMc')

    def make_Ek_grid(self):
        """Make an energy grid on which 2vN equations are solved. """
        if self.funcp.kpnt is None:
            raise ValueError('kpnt needs to be specified.')
        if self.si.nleads > 0:
            dmin = np.min(self.leads.dlst)
            dmax = np.max(self.leads.dlst)
            Ek_grid, kpnt = self.Ek_grid, self.funcp.kpnt
            if Ek_grid[0] != dmin or Ek_grid[-1] != dmax or Ek_grid.shape[0] != kpnt:
                self.funcp.dmin = dmin
                self.funcp.dmax = dmax
                self.Ek_grid = np.linspace(dmin, dmax, kpnt)
                #
                if self.niter != -1:
                    print("WARNING: Ek_grid has changed. Restarting the calculation.")
                    self.restart()
                #
                if ((dmin*np.ones(self.si.nleads)).tolist() != self.leads.dlst.T[0].tolist() or
                    (dmax*np.ones(self.si.nleads)).tolist() != self.leads.dlst.T[1].tolist()):
                    print("WARNING: The bandwidth and Ek_grid for all leads will be the same: from "+
                          "dmin="+str(dmin)+" to dmax="+str(dmax)+".")

    def restart(self):
        """Restart values of some variables for new calculations."""
        self.kern, self.bvec = None, None
        self.sol0, self.phi0, self.phi1 = None, None, None
        self.current = None
        self.energy_current = None
        self.heat_current = None
        #
        self.niter = -1
        self.phi1k = None
        self.phi1k_delta = None
        self.hphi1k_delta = None
        self.kern1k = None
        #
        self.iters = []

    def iteration(self):
        """Makes one iteration for solution of the 2vN integral equation."""
        self.iterate(self)
        self.phi1k = self.phi1k_delta if self.phi1k is None else self.phi1k + self.phi1k_delta
        self.get_phi1_phi0(self)
        self.niter += 1
        self.kern_phi0(self)
        self.solve_kern()
        self.generate_current(self)
        #
        self.iters.append(Iterations2vN(self))


    def solve(self, qdq=True, rotateq=True, masterq=True, restartq=True,
                    niter=None, func_iter=None, *args, **kwargs):
        """
        Solves the 2vN approach integral equations iteratively.

        Parameters
        ----------
        qdq : bool
            Diagonalise many-body quantum dot Hamiltonian
            and express the lead matrix Tba in the eigenbasis.
        rotateq : bool
            Rotate the many-body tunneling matrix Tba.
        masterq : bool
            Solve the master equation.
        restartq : bool
            Call restart() and erase old values of variables.
            To continue from the last iteration set restarq=False.
        niter : int
            Number of iterations to perform.
        func_iter : function
            User defined function which is performed after every iteration and
            takes Approach2vN object as an input.
        """
        if restartq:
            self.restart()
        if qdq:
            self.qd.diagonalise()
            if rotateq:
                self.leads.rotate(self.qd.vecslst)
        #
        if masterq:
            # Exception
            if niter is None:
                raise ValueError('Number of iterations niter needs to be specified')
            self.make_Ek_grid()
            #
            for it in range(niter):
                self.iteration()
                if not func_iter is None:
                    func_iter(self)



"""Module containing python functions, which generate first order Pauli kernel."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import numpy as np
import itertools

from ..mytypes import complexnp
from ..mytypes import doublenp

from ..specfunc import func_pauli
from ..aprclass import Approach

def generate_paulifct(sys):
    """
    Make factors used for generating Pauli master equation kernel.

    Parameters
    ----------
    sys : Approach
        Approach object.

    Modifies:
    sys.paulifct : array
        Factors used for generating Pauli master equation kernel.
    """
    (E, Tba, si, mulst, tlst, dlst) = (sys.qd.Ea, sys.leads.Tba, sys.si,
                                       sys.leads.mulst, sys.leads.tlst, sys.leads.dlst)
    itype = sys.funcp.itype
    paulifct = np.zeros((si.nleads, si.ndm1, 2), dtype=doublenp)
    for charge in range(si.ncharge-1):
        ccharge = charge+1
        bcharge = charge
        for c, b in itertools.product(si.statesdm[ccharge], si.statesdm[bcharge]):
            cb = si.get_ind_dm1(c, b, bcharge)
            Ecb = E[c]-E[b]
            for l in range(si.nleads):
                xcb = (Tba[l, b, c]*Tba[l, c, b]).real
                rez = func_pauli(Ecb, mulst[l], tlst[l], dlst[l,0], dlst[l,1], itype)
                paulifct[l, cb, 0] = xcb*rez[0]
                paulifct[l, cb, 1] = xcb*rez[1]
    sys.paulifct = paulifct
    return 0

#---------------------------------------------------------------------------------------------------
# Pauli master equation
#---------------------------------------------------------------------------------------------------
def generate_kern_pauli(sys):
    """
    Generate Pauli master equation kernel.

    Parameters
    ----------
    sys : Approach
        Approach object.

    Modifies:
    sys.kern : array
        Kernel matrix for Pauli master equation.
    sys.bvec : array
        Right hand side column vector for master equation.
        The entry funcp.norm_row is 1 representing normalization condition.
    """
    (paulifct, si, symq, norm_rowp) = (sys.paulifct, sys.si, sys.funcp.symq, sys.funcp.norm_row)
    norm_row = norm_rowp if symq else si.npauli
    last_row = si.npauli-1 if symq else si.npauli
    kern = np.zeros((last_row+1, si.npauli), dtype=doublenp)
    bvec = np.zeros(last_row+1, dtype=doublenp)
    bvec[norm_row] = 1
    for charge in range(si.ncharge):
        for b in si.statesdm[charge]:
            bb = si.get_ind_dm0(b, b, charge)
            bb_bool = si.get_ind_dm0(b, b, charge, 2)
            kern[norm_row, bb] += 1
            if not (symq and bb == norm_row) and bb_bool:
                for a in si.statesdm[charge-1]:
                    aa = si.get_ind_dm0(a, a, charge-1)
                    ba = si.get_ind_dm1(b, a, charge-1)
                    for l in range(si.nleads):
                        kern[bb, bb] -= paulifct[l, ba, 1]
                        kern[bb, aa] += paulifct[l, ba, 0]
                for c in si.statesdm[charge+1]:
                    cc = si.get_ind_dm0(c, c, charge+1)
                    cb = si.get_ind_dm1(c, b, charge)
                    for l in range(si.nleads):
                        kern[bb, bb] -= paulifct[l, cb, 0]
                        kern[bb, cc] += paulifct[l, cb, 1]
    sys.kern = kern
    sys.bvec = bvec
    return 0

def generate_current_pauli(sys):
    """
    Calculates currents using Pauli master equation approach.

    Parameters
    ----------
    sys : Approach
        Approach object.

    Modifies:
    sys.current : array
        Values of the current having nleads entries.
    sys.energy_current : array
        Values of the energy current having nleads entries.
    sys.heat_current : array
        Values of the heat current having nleads entries.
    """
    (phi0, E, paulifct, si) = (sys.phi0, sys.qd.Ea, sys.paulifct, sys.si)
    current = np.zeros(si.nleads, dtype=doublenp)
    energy_current = np.zeros(si.nleads, dtype=doublenp)
    for charge in range(si.ncharge-1):
        ccharge = charge+1
        bcharge = charge
        for c in si.statesdm[ccharge]:
            cc = si.get_ind_dm0(c, c, ccharge)
            for b in si.statesdm[bcharge]:
                bb = si.get_ind_dm0(b, b, bcharge)
                cb = si.get_ind_dm1(c, b, bcharge)
                for l in range(si.nleads):
                    fct1 = +phi0[bb]*paulifct[l, cb, 0]
                    fct2 = -phi0[cc]*paulifct[l, cb, 1]
                    current[l] += fct1 + fct2
                    energy_current[l] += -(E[b]-E[c])*(fct1 + fct2)
    sys.current = current
    sys.energy_current = energy_current
    sys.heat_current = energy_current - current*sys.leads.mulst
    return 0

def generate_vec_pauli(phi0, sys):
    """
    Acts on given phi0 with Liouvillian of Pauli approach.

    Parameters
    ----------
    phi0 : array
        Some values of zeroth order density matrix elements.
    sys : Approach
        Approach object.

    Returns
    -------
    dphi0_dt : array
        Values of zeroth order density matrix elements
        after acting with Liouvillian, i.e., dphi0_dt=L(phi0p).
    """
    (paulifct, si, norm_row) = (sys.paulifct, sys.si, sys.funcp.norm_row)
    dphi0_dt = np.zeros(si.npauli, dtype=doublenp)
    norm = 0
    for charge in range(si.ncharge):
        for b in si.statesdm[charge]:
            bb = si.get_ind_dm0(b, b, charge)
            bb_bool = si.get_ind_dm0(b, b, charge, 2)
            norm += phi0[bb]
            if bb_bool:
                for a in si.statesdm[charge-1]:
                    aa = si.get_ind_dm0(a, a, charge-1)
                    ba = si.get_ind_dm1(b, a, charge-1)
                    for l in range(si.nleads):
                        dphi0_dt[bb] -= paulifct[l, ba, 1]*phi0[bb]
                        dphi0_dt[bb] += paulifct[l, ba, 0]*phi0[aa]
                for c in si.statesdm[charge+1]:
                    cc = si.get_ind_dm0(c, c, charge+1)
                    cb = si.get_ind_dm1(c, b, charge)
                    for l in range(si.nleads):
                        dphi0_dt[bb] -= paulifct[l, cb, 0]*phi0[bb]
                        dphi0_dt[bb] += paulifct[l, cb, 1]*phi0[cc]
    dphi0_dt[norm_row] = norm-1
    return dphi0_dt

class Approach_pyPauli(Approach):

    kerntype = 'pyPauli'
    generate_fct = staticmethod(generate_paulifct)
    generate_kern = staticmethod(generate_kern_pauli)
    generate_current = staticmethod(generate_current_pauli)
    generate_vec = staticmethod(generate_vec_pauli)
#---------------------------------------------------------------------------------------------------



"""Module containing pcython functions, which generate first order Redfield kernel.
   For docstrings see documentation of module neumann1."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import numpy as np
import itertools

from ..mytypes import doublenp
from ..mytypes import complexnp

from ..aprclass import Approach
from .neumann1 import generate_phi1fct

#---------------------------------------------------------------------------------------------------
# Redfield approach
#---------------------------------------------------------------------------------------------------
def generate_kern_redfield(sys):
    (E, Tba, phi1fct, si, symq, norm_rowp) = (sys.qd.Ea, sys.leads.Tba, sys.phi1fct, sys.si,
                                              sys.funcp.symq, sys.funcp.norm_row)
    norm_row = norm_rowp if symq else si.ndm0r
    last_row = si.ndm0r-1 if symq else si.ndm0r
    kern = np.zeros((last_row+1, si.ndm0r), dtype=doublenp)
    bvec = np.zeros(last_row+1, dtype=doublenp)
    bvec[norm_row] = 1
    npauli, ndm0, nleads = si.npauli, si.ndm0, si.nleads
    for charge in range(si.ncharge):
        acharge = charge-1
        bcharge = charge
        ccharge = charge+1
        for b, bp in itertools.combinations_with_replacement(si.statesdm[bcharge], 2):
            bbp = si.get_ind_dm0(b, bp, bcharge)
            bbp_bool = si.get_ind_dm0(b, bp, bcharge, 2)
            if bbp != -1 and bbp_bool:
                bbpi = ndm0 + bbp - npauli
                bbpi_bool = True if bbpi >= ndm0 else False
                if bbpi_bool:
                    kern[bbp, bbpi] += E[b]-E[bp]
                    kern[bbpi, bbp] += E[bp]-E[b]
                #--------------------------------------------------
                for a, ap in itertools.product(si.statesdm[acharge], si.statesdm[acharge]):
                    aap = si.get_ind_dm0(a, ap, acharge)
                    if aap != -1:
                        bpap = si.get_ind_dm1(bp, ap, acharge)
                        ba = si.get_ind_dm1(b, a, acharge)
                        fct_aap = 0
                        for l in range(nleads):
                            fct_aap += (+Tba[l, b, a]*Tba[l, ap, bp]*phi1fct[l, bpap, 0].conjugate()
                                        -Tba[l, b, a]*Tba[l, ap, bp]*phi1fct[l, ba, 0])
                        aapi = ndm0 + aap - npauli
                        aap_sgn = +1 if si.get_ind_dm0(a, ap, acharge, maptype=3) else -1
                        kern[bbp, aap] += fct_aap.imag
                        if aapi >= ndm0:
                            kern[bbp, aapi] += fct_aap.real*aap_sgn
                            if bbpi_bool:
                                kern[bbpi, aapi] += fct_aap.imag*aap_sgn
                        if bbpi_bool:
                            kern[bbpi, aap] -= fct_aap.real
                #--------------------------------------------------
                for bpp in si.statesdm[bcharge]:
                    bppbp = si.get_ind_dm0(bpp, bp, bcharge)
                    if bppbp != -1:
                        fct_bppbp = 0
                        for a in si.statesdm[acharge]:
                            bppa = si.get_ind_dm1(bpp, a, acharge)
                            for l in range(nleads):
                                fct_bppbp += +Tba[l, b, a]*Tba[l, a, bpp]*phi1fct[l, bppa, 1].conjugate()
                        for c in si.statesdm[ccharge]:
                            cbpp = si.get_ind_dm1(c, bpp, bcharge)
                            for l in range(nleads):
                                fct_bppbp += +Tba[l, b, c]*Tba[l, c, bpp]*phi1fct[l, cbpp, 0]
                        bppbpi = ndm0 + bppbp - npauli
                        bppbp_sgn = +1 if si.get_ind_dm0(bpp, bp, bcharge, maptype=3) else -1
                        kern[bbp, bppbp] += fct_bppbp.imag
                        if bppbpi >= ndm0:
                            kern[bbp, bppbpi] += fct_bppbp.real*bppbp_sgn
                            if bbpi_bool:
                                kern[bbpi, bppbpi] += fct_bppbp.imag*bppbp_sgn
                        if bbpi_bool:
                            kern[bbpi, bppbp] -= fct_bppbp.real
                    #--------------------------------------------------
                    bbpp = si.get_ind_dm0(b, bpp, bcharge)
                    if bbpp != -1:
                        fct_bbpp = 0
                        for a in si.statesdm[acharge]:
                            bppa = si.get_ind_dm1(bpp, a, acharge)
                            for l in range(nleads):
                                fct_bbpp += -Tba[l, bpp, a]*Tba[l, a, bp]*phi1fct[l, bppa, 1]
                        for c in si.statesdm[ccharge]:
                            cbpp = si.get_ind_dm1(c, bpp, bcharge)
                            for l in range(nleads):
                                fct_bbpp += -Tba[l, bpp, c]*Tba[l, c, bp]*phi1fct[l, cbpp, 0].conjugate()
                        bbppi = ndm0 + bbpp - npauli
                        bbpp_sgn = +1 if si.get_ind_dm0(b, bpp, bcharge, maptype=3) else -1
                        kern[bbp, bbpp] += fct_bbpp.imag
                        if bbppi >= ndm0:
                            kern[bbp, bbppi] += fct_bbpp.real*bbpp_sgn
                            if bbpi_bool:
                                kern[bbpi, bbppi] += fct_bbpp.imag*bbpp_sgn
                        if bbpi_bool:
                            kern[bbpi, bbpp] -= fct_bbpp.real
                #--------------------------------------------------
                for c, cp in itertools.product(si.statesdm[ccharge], si.statesdm[ccharge]):
                    ccp = si.get_ind_dm0(c, cp, ccharge)
                    if ccp != -1:
                        cpbp = si.get_ind_dm1(cp, bp, bcharge)
                        cb = si.get_ind_dm1(c, b, bcharge)
                        fct_ccp = 0
                        for l in range(nleads):
                            fct_ccp += (+Tba[l, b, c]*Tba[l, cp, bp]*phi1fct[l, cpbp, 1]
                                        -Tba[l, b, c]*Tba[l, cp, bp]*phi1fct[l, cb, 1].conjugate())
                        ccpi = ndm0 + ccp - npauli
                        ccp_sgn = +1 if si.get_ind_dm0(c, cp, ccharge, maptype=3) else -1
                        kern[bbp, ccp] += fct_ccp.imag
                        if ccpi >= ndm0:
                            kern[bbp, ccpi] += fct_ccp.real*ccp_sgn
                            if bbpi_bool:
                                kern[bbpi, ccpi] += fct_ccp.imag*ccp_sgn
                        if bbpi_bool:
                            kern[bbpi, ccp] -= fct_ccp.real
                #--------------------------------------------------
    # Normalisation condition
    kern[norm_row] = np.zeros(si.ndm0r, dtype=doublenp)
    for charge in range(si.ncharge):
        for b in si.statesdm[charge]:
            bb = si.get_ind_dm0(b, b, charge)
            kern[norm_row, bb] += 1
    sys.kern = kern
    sys.bvec = bvec
    return 0

def generate_current_redfield(sys):
    (phi0p, E, Tba, phi1fct, phi1fct_energy, si) = (sys.phi0, sys.qd.Ea, sys.leads.Tba,
                                                    sys.phi1fct, sys.phi1fct_energy, sys.si)
    phi1 = np.zeros((si.nleads, si.ndm1), dtype=complexnp)
    current = np.zeros(si.nleads, dtype=complexnp)
    energy_current = np.zeros(si.nleads, dtype=complexnp)
    #
    phi0 = np.zeros(si.ndm0, dtype=complexnp)
    phi0[0:si.npauli] = phi0p[0:si.npauli]
    phi0[si.npauli:si.ndm0] = phi0p[si.npauli:si.ndm0] + 1j*phi0p[si.ndm0:]
    #
    nleads = si.nleads
    for charge in range(si.ncharge-1):
        ccharge = charge+1
        bcharge = charge
        for c, b in itertools.product(si.statesdm[ccharge], si.statesdm[bcharge]):
            cb = si.get_ind_dm1(c, b, bcharge)
            for l in range(nleads):
                for bp in si.statesdm[bcharge]:
                    bpb = si.get_ind_dm0(bp, b, bcharge)
                    if bpb != -1:
                        cbp = si.get_ind_dm1(c, bp, bcharge)
                        fct1 = phi1fct[l, cbp, 0]
                        fct1h = phi1fct_energy[l, cbp, 0]
                        bpb_conj = si.get_ind_dm0(bp, b, bcharge, maptype=3)
                        phi0bpb = phi0[bpb] if bpb_conj else phi0[bpb].conjugate()
                        phi1[l, cb] += Tba[l, c, bp]*phi0bpb*fct1
                        current[l] += Tba[l, b, c]*Tba[l, c, bp]*phi0bpb*fct1
                        energy_current[l] += Tba[l, b, c]*Tba[l, c, bp]*phi0bpb*fct1h
                for cp in si.statesdm[ccharge]:
                    ccp = si.get_ind_dm0(c, cp, ccharge)
                    if ccp != -1:
                        cpb = si.get_ind_dm1(cp, b, bcharge)
                        fct2 = phi1fct[l, cpb, 1]
                        fct2h = phi1fct_energy[l, cpb, 1]
                        ccp_conj = si.get_ind_dm0(c, cp, ccharge, maptype=3)
                        phi0ccp = phi0[ccp] if ccp_conj else phi0[ccp].conjugate()
                        phi1[l, cb] += Tba[l, cp, b]*phi0ccp*fct2
                        current[l] += Tba[l, b, c]*phi0ccp*Tba[l, cp, b]*fct2
                        energy_current[l] += Tba[l, b, c]*phi0ccp*Tba[l, cp, b]*fct2h
        sys.phi1 = phi1
    sys.current = np.array(-2*current.imag, dtype=doublenp)
    sys.energy_current = np.array(-2*energy_current.imag, dtype=doublenp)
    sys.heat_current = sys.energy_current - sys.current*sys.leads.mulst
    return 0

def generate_vec_redfield(phi0p, sys):
    (E, Tba, phi1fct, si, norm_row) = (sys.qd.Ea, sys.leads.Tba, sys.phi1fct,
                                       sys.si, sys.funcp.norm_row)
    #
    phi0 = np.zeros(si.ndm0, dtype=complexnp)
    phi0[0:si.npauli] = phi0p[0:si.npauli]
    phi0[si.npauli:si.ndm0] = phi0p[si.npauli:si.ndm0] + 1j*phi0p[si.ndm0:]
    #
    i_dphi0_dt = np.zeros(si.ndm0, dtype=complexnp)
    norm = 0
    nleads = si.nleads
    for charge in range(si.ncharge):
        acharge = charge-1
        bcharge = charge
        ccharge = charge+1
        for b, bp in itertools.combinations_with_replacement(si.statesdm[bcharge], 2):
            bbp = si.get_ind_dm0(b, bp, bcharge)
            if bbp != -1:
                if b == bp: norm += phi0[bbp]
                bbp_bool = si.get_ind_dm0(b, bp, bcharge, maptype=2)
                if bbp_bool:
                    i_dphi0_dt[bbp] += (E[b]-E[bp])*phi0[bbp]
                    #--------------------------------------------------
                    for a, ap in itertools.product(si.statesdm[acharge], si.statesdm[acharge]):
                        aap = si.get_ind_dm0(a, ap, acharge)
                        if aap != -1:
                            bpap = si.get_ind_dm1(bp, ap, acharge)
                            ba = si.get_ind_dm1(b, a, acharge)
                            fct_aap = 0
                            for l in range(nleads):
                                fct_aap += (+Tba[l, b, a]*Tba[l, ap, bp]*phi1fct[l, bpap, 0].conjugate()
                                            -Tba[l, b, a]*Tba[l, ap, bp]*phi1fct[l, ba, 0])
                            phi0aap = ( phi0[aap] if si.get_ind_dm0(a, ap, acharge, maptype=3)
                                                  else phi0[aap].conjugate() )
                            i_dphi0_dt[bbp] += fct_aap*phi0aap
                    #--------------------------------------------------
                    for bpp in si.statesdm[bcharge]:
                        bppbp = si.get_ind_dm0(bpp, bp, bcharge)
                        if bppbp != -1:
                            fct_bppbp = 0
                            for a in si.statesdm[acharge]:
                                bppa = si.get_ind_dm1(bpp, a, acharge)
                                for l in range(nleads):
                                    fct_bppbp += +Tba[l, b, a]*Tba[l, a, bpp]*phi1fct[l, bppa, 1].conjugate()
                            for c in si.statesdm[ccharge]:
                                cbpp = si.get_ind_dm1(c, bpp, bcharge)
                                for l in range(nleads):
                                    fct_bppbp += +Tba[l, b, c]*Tba[l, c, bpp]*phi1fct[l, cbpp, 0]
                            phi0bppbp = ( phi0[bppbp] if si.get_ind_dm0(bpp, bp, bcharge, maptype=3)
                                                      else  phi0[bppbp].conjugate() )
                            i_dphi0_dt[bbp] += fct_bppbp*phi0bppbp
                        #--------------------------------------------------
                        bbpp = si.get_ind_dm0(b, bpp, charge)
                        if bbpp != -1:
                            fct_bbpp = 0
                            for a in si.statesdm[acharge]:
                                bppa = si.get_ind_dm1(bpp, a, acharge)
                                for l in range(nleads):
                                    fct_bbpp += -Tba[l, bpp, a]*Tba[l, a, bp]*phi1fct[l, bppa, 1]
                            for c in si.statesdm[ccharge]:
                                cbpp = si.get_ind_dm1(c, bpp, bcharge)
                                for l in range(nleads):
                                    fct_bbpp += -Tba[l, bpp, c]*Tba[l, c, bp]*phi1fct[l, cbpp, 0].conjugate()
                            phi0bbpp = ( phi0[bbpp] if si.get_ind_dm0(b, bpp, bcharge, maptype=3)
                                                    else phi0[bbpp].conjugate() )
                            i_dphi0_dt[bbp] += fct_bbpp*phi0bbpp
                    #--------------------------------------------------
                    for c, cp in itertools.product(si.statesdm[ccharge], si.statesdm[ccharge]):
                        ccp = si.get_ind_dm0(c, cp, ccharge)
                        if ccp != -1:
                            cpbp = si.get_ind_dm1(cp, bp, bcharge)
                            cb = si.get_ind_dm1(c, b, bcharge)
                            fct_ccp = 0
                            for l in range(nleads):
                                fct_ccp += (+Tba[l, b, c]*Tba[l, cp, bp]*phi1fct[l, cpbp, 1]
                                            -Tba[l, b, c]*Tba[l, cp, bp]*phi1fct[l, cb, 1].conjugate())
                            phi0ccp = ( phi0[ccp] if si.get_ind_dm0(c, cp, ccharge, maptype=3)
                                                  else phi0[ccp].conjugate() )
                            i_dphi0_dt[bbp] += fct_ccp*phi0ccp
                    #--------------------------------------------------
    i_dphi0_dt[norm_row] = 1j*(norm-1)
    return np.concatenate((i_dphi0_dt.imag, i_dphi0_dt[si.npauli:si.ndm0].real))

class Approach_pyRedfield(Approach):

    kerntype = 'pyRedfield'
    generate_fct = staticmethod(generate_phi1fct)
    generate_kern = staticmethod(generate_kern_redfield)
    generate_current = staticmethod(generate_current_redfield)
    generate_vec = staticmethod(generate_vec_redfield)
#---------------------------------------------------------------------------------------------------
